# Copyright 2015 Reliance Jio Infocomm Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import collections
from unittest import mock

import fixtures
from keystoneauth1 import exceptions
from oslotest import base
import testscenarios.testcase

from ceilometer.objectstore import rgw
from ceilometer.objectstore import rgw_client
from ceilometer.polling import manager
from ceilometer import service

bucket_list1 = [rgw_client.RGWAdminClient.Bucket('somefoo1', 10, 7)]
bucket_list2 = [rgw_client.RGWAdminClient.Bucket('somefoo2', 2, 9)]
bucket_list3 = [rgw_client.RGWAdminClient.Bucket('unlisted', 100, 100)]

GET_BUCKETS = [('tenant-000', {'num_buckets': 2, 'size': 1042,
                               'num_objects': 1001, 'buckets': bucket_list1}),
               ('tenant-001', {'num_buckets': 2, 'size': 1042,
                               'num_objects': 1001, 'buckets': bucket_list2}),
               ('tenant-002-ignored', {'num_buckets': 2, 'size': 1042,
                                       'num_objects': 1001,
                                       'buckets': bucket_list3})]

GET_USAGE = [('tenant-000', 10),
             ('tenant-001', 11),
             ('tenant-002-ignored', 12)]

Tenant = collections.namedtuple('Tenant', 'id')
ASSIGNED_TENANTS = [Tenant('tenant-000'), Tenant('tenant-001')]


class TestManager(manager.AgentManager):

    def __init__(self, worker_id, conf):
        super().__init__(worker_id, conf)
        self._keystone = mock.Mock()
        self._catalog = (self._keystone.session.auth.get_access.
                         return_value.service_catalog)
        self._catalog.url_for.return_value = 'http://foobar/endpoint'


class TestRgwPollster(testscenarios.testcase.WithScenarios,
                      base.BaseTestCase):

    # Define scenarios to run all of the tests against all of the
    # pollsters.
    scenarios = [
        ('radosgw.objects',
         {'factory': rgw.ObjectsPollster}),
        ('radosgw.objects.size',
         {'factory': rgw.ObjectsSizePollster}),
        ('radosgw.objects.containers',
         {'factory': rgw.ObjectsContainersPollster}),
        ('radosgw.containers.objects',
         {'factory': rgw.ContainersObjectsPollster}),
        ('radosgw.containers.objects.size',
         {'factory': rgw.ContainersSizePollster}),
        ('radosgw.api.request',
         {'factory': rgw.UsagePollster}),
    ]

    @staticmethod
    def fake_ks_service_catalog_url_for(*args, **kwargs):
        raise exceptions.EndpointNotFound("Fake keystone exception")

    def fake_iter_accounts(self, ksclient, cache, tenants):
        tenant_ids = [t.id for t in tenants]
        for i in self.ACCOUNTS:
            if i[0] in tenant_ids:
                yield i

    def setUp(self):
        super().setUp()
        conf = service.prepare_service([], [])
        conf.set_override('radosgw', 'object-store',
                          group='service_types')
        self.pollster = self.factory(conf)
        self.manager = TestManager(0, conf)

        if self.pollster.CACHE_KEY_METHOD == 'rgw.get_bucket':
            self.ACCOUNTS = GET_BUCKETS
        else:
            self.ACCOUNTS = GET_USAGE

    def tearDown(self):
        super().tearDown()
        rgw._Base._ENDPOINT = None

    def test_iter_accounts_no_cache(self):
        cache = {}
        with fixtures.MockPatchObject(self.factory, '_get_account_info',
                                      return_value=[]):
            data = list(self.pollster._iter_accounts(mock.Mock(), cache,
                                                     ASSIGNED_TENANTS))

        self.assertIn(self.pollster.CACHE_KEY_METHOD, cache)
        self.assertEqual([], data)

    def test_iter_accounts_cached(self):
        # Verify that if a method has already been called, _iter_accounts
        # uses the cached version and doesn't call rgw_clinet.
        mock_method = mock.Mock()
        mock_method.side_effect = AssertionError(
            'should not be called',
        )

        api_method = 'get_%s' % self.pollster.METHOD

        with fixtures.MockPatchObject(rgw_client.RGWAdminClient,
                                      api_method, new=mock_method):
            cache = {self.pollster.CACHE_KEY_METHOD: [self.ACCOUNTS[0]]}
            data = list(self.pollster._iter_accounts(mock.Mock(), cache,
                                                     ASSIGNED_TENANTS))
        self.assertEqual([self.ACCOUNTS[0]], data)

    def test_metering(self):
        with fixtures.MockPatchObject(self.factory, '_iter_accounts',
                                      side_effect=self.fake_iter_accounts):
            samples = list(self.pollster.get_samples(self.manager, {},
                                                     ASSIGNED_TENANTS))

        self.assertEqual(2, len(samples), self.pollster.__class__)

    def test_get_meter_names(self):
        with fixtures.MockPatchObject(self.factory, '_iter_accounts',
                                      side_effect=self.fake_iter_accounts):
            samples = list(self.pollster.get_samples(self.manager, {},
                                                     ASSIGNED_TENANTS))

        self.assertEqual({samples[0].name},
                         {s.name for s in samples})

    def test_only_poll_assigned(self):
        mock_method = mock.MagicMock()
        endpoint = 'http://127.0.0.1:8000/admin'
        api_method = 'get_%s' % self.pollster.METHOD
        with fixtures.MockPatchObject(rgw_client.RGWAdminClient,
                                      api_method, new=mock_method):
            with fixtures.MockPatchObject(
                    self.manager._catalog, 'url_for',
                    return_value=endpoint):
                list(self.pollster.get_samples(self.manager, {},
                                               ASSIGNED_TENANTS))
        expected = [mock.call(t.id)
                    for t in ASSIGNED_TENANTS]
        self.assertEqual(expected, mock_method.call_args_list)

    def test_get_endpoint_only_once(self):
        mock_url_for = mock.MagicMock()
        mock_url_for.return_value = '/endpoint'
        api_method = 'get_%s' % self.pollster.METHOD
        with fixtures.MockPatchObject(rgw_client.RGWAdminClient, api_method,
                                      new=mock.MagicMock()):
            with fixtures.MockPatchObject(
                    self.manager._catalog, 'url_for',
                    new=mock_url_for):
                list(self.pollster.get_samples(self.manager, {},
                                               ASSIGNED_TENANTS))
                list(self.pollster.get_samples(self.manager, {},
                                               ASSIGNED_TENANTS))
        self.assertEqual(1, mock_url_for.call_count)

    def test_endpoint_notfound(self):
        with fixtures.MockPatchObject(
                self.manager._catalog, 'url_for',
                side_effect=self.fake_ks_service_catalog_url_for):
            samples = list(self.pollster.get_samples(self.manager, {},
                                                     ASSIGNED_TENANTS))

        self.assertEqual(0, len(samples))



#
# Copyright 2012 Red Hat, Inc
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
"""Inspector abstraction for read-only access to hypervisors."""

import collections

from oslo_config import cfg
from oslo_log import log
from stevedore import driver

import ceilometer


OPTS = [
    cfg.StrOpt('hypervisor_inspector',
               default='libvirt',
               choices=['libvirt'],
               deprecated_for_removal=True,
               deprecated_reason='libvirt is the only supported hypervisor',
               help='Inspector to use for inspecting the hypervisor layer.')
]


LOG = log.getLogger(__name__)


# Named tuple representing instance statistics

class InstanceStats:
    fields = [
        'power_state',       # the power state of the domain
        'cpu_number',        # number: number of CPUs
        'cpu_time',          # time: cumulative CPU time
        'memory_actual',     # actual: Amount of allocated memory
        'memory_available',  # available: Amount of usable memory
        'memory_usage',      # usage: Amount of memory used
        'memory_resident',   #
        'memory_swap_in',    # memory swap in
        'memory_swap_out',   # memory swap out
        'cpu_cycles',        # cpu_cycles: the number of cpu cycles one
                             #   instruction needs
        'instructions',      # instructions: the count of instructions
        'cache_references',  # cache_references: the count of cache hits
        'cache_misses',      # cache_misses: the count of caches misses
    ]

    def __init__(self, **kwargs):
        for k in self.fields:
            setattr(self, k, kwargs.pop(k, None))
        if kwargs:
            raise AttributeError(
                "'InstanceStats' object has no attributes '%s'" % kwargs)


# Named tuple representing vNIC statistics.
#
# name: the name of the vNIC
# mac: the MAC address
# fref: the filter ref
# parameters: miscellaneous parameters
# rx_bytes: number of received bytes
# rx_packets: number of received packets
# tx_bytes: number of transmitted bytes
# tx_packets: number of transmitted packets
#
InterfaceStats = collections.namedtuple('InterfaceStats',
                                        ['name', 'mac', 'fref', 'parameters',
                                         'rx_bytes', 'tx_bytes',
                                         'rx_packets', 'tx_packets',
                                         'rx_drop', 'tx_drop',
                                         'rx_errors', 'tx_errors',
                                         'rx_bytes_delta', 'tx_bytes_delta'])


# Named tuple representing vNIC rate statistics.
#
# name: the name of the vNIC
# mac: the MAC address
# fref: the filter ref
# parameters: miscellaneous parameters
# rx_bytes_rate: rate of received bytes
# tx_bytes_rate: rate of transmitted bytes
#
InterfaceRateStats = collections.namedtuple('InterfaceRateStats',
                                            ['name', 'mac',
                                             'fref', 'parameters',
                                             'rx_bytes_rate', 'tx_bytes_rate'])


# Named tuple representing disk statistics.
#
# read_bytes: number of bytes read
# read_requests: number of read operations
# write_bytes: number of bytes written
# write_requests: number of write operations
# errors: number of errors
#
DiskStats = collections.namedtuple('DiskStats',
                                   ['device',
                                    'read_bytes', 'read_requests',
                                    'write_bytes', 'write_requests',
                                    'errors', 'wr_total_times',
                                    'rd_total_times'])

# Named tuple representing disk rate statistics.
#
# read_bytes_rate: number of bytes read per second
# read_requests_rate: number of read operations per second
# write_bytes_rate: number of bytes written per second
# write_requests_rate: number of write operations per second
#
DiskRateStats = collections.namedtuple('DiskRateStats',
                                       ['device',
                                        'read_bytes_rate',
                                        'read_requests_rate',
                                        'write_bytes_rate',
                                        'write_requests_rate'])

# Named tuple representing disk Information.
#
# capacity: capacity of the disk
# allocation: allocation of the disk
# physical: usage of the disk

DiskInfo = collections.namedtuple('DiskInfo',
                                  ['device',
                                   'capacity',
                                   'allocation',
                                   'physical'])


# Exception types
#
class InspectorException(Exception):
    def __init__(self, message=None):
        super().__init__(message)


class InstanceNotFoundException(InspectorException):
    pass


class InstanceShutOffException(InspectorException):
    pass


class NoDataException(InspectorException):
    pass


# Main virt inspector abstraction layering over the hypervisor API.
#
class Inspector:

    def __init__(self, conf):
        self.conf = conf

    def inspect_instance(self, instance, duration):
        """Inspect the CPU statistics for an instance.

        :param instance: the target instance
        :param duration: the last 'n' seconds, over which the value should be
               inspected
        :return: the instance stats
        """
        raise ceilometer.NotImplementedError

    def inspect_vnics(self, instance, duration):
        """Inspect the vNIC statistics for an instance.

        :param instance: the target instance
        :param duration: the last 'n' seconds, over which the value should be
               inspected
        :return: for each vNIC, the number of bytes & packets
                 received and transmitted
        """
        raise ceilometer.NotImplementedError

    def inspect_vnic_rates(self, instance, duration):
        """Inspect the vNIC rate statistics for an instance.

        :param instance: the target instance
        :param duration: the last 'n' seconds, over which the value should be
               inspected
        :return: for each vNIC, the rate of bytes & packets
                 received and transmitted
        """
        raise ceilometer.NotImplementedError

    def inspect_disks(self, instance, duration):
        """Inspect the disk statistics for an instance.

        :param instance: the target instance
        :param duration: the last 'n' seconds, over which the value should be
               inspected
        :return: for each disk, the number of bytes & operations
                 read and written, and the error count
        """
        raise ceilometer.NotImplementedError

    def inspect_disk_rates(self, instance, duration):
        """Inspect the disk statistics as rates for an instance.

        :param instance: the target instance
        :param duration: the last 'n' seconds, over which the value should be
               inspected
        :return: for each disk, the number of bytes & operations
                 read and written per second, with the error count
        """
        raise ceilometer.NotImplementedError

    def inspect_disk_info(self, instance, duration):
        """Inspect the disk information for an instance.

        :param instance: the target instance
        :param duration: the last 'n' seconds, over which the value should be
               inspected
        :return: for each disk , capacity , allocation and usage
        """
        raise ceilometer.NotImplementedError


def get_hypervisor_inspector(conf):
    try:
        namespace = 'ceilometer.compute.virt'
        mgr = driver.DriverManager(namespace,
                                   conf.hypervisor_inspector,
                                   invoke_on_load=True,
                                   invoke_args=(conf, ))
        return mgr.driver
    except ImportError as e:
        LOG.error("Unable to load the hypervisor inspector: %s", e)
        return Inspector(conf)



#
# Copyright 2012 eNovance <licensing@enovance.com>
# Copyright 2012 Red Hat, Inc
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import time

from ceilometer.compute.pollsters import instance_stats
from ceilometer.compute.virt import inspector as virt_inspector
from ceilometer.polling import manager
from ceilometer.tests.unit.compute.pollsters import base


class TestCPUPollster(base.TestPollsterBase):

    def test_get_samples(self):
        self._mock_inspect_instance(
            virt_inspector.InstanceStats(cpu_time=1 * (10 ** 6), cpu_number=2),
            virt_inspector.InstanceStats(cpu_time=3 * (10 ** 6), cpu_number=2),
            # cpu_time resets on instance restart
            virt_inspector.InstanceStats(cpu_time=2 * (10 ** 6), cpu_number=2),
        )

        mgr = manager.AgentManager(0, self.CONF)
        pollster = instance_stats.CPUPollster(self.CONF)

        def _verify_cpu_metering(expected_time):
            cache = {}
            samples = list(pollster.get_samples(mgr, cache, [self.instance]))
            self.assertEqual(1, len(samples))
            self.assertEqual({'cpu'}, {s.name for s in samples})
            self.assertEqual(expected_time, samples[0].volume)
            self.assertEqual(2, samples[0].resource_metadata.get('cpu_number'))
            # ensure elapsed time between polling cycles is non-zero
            time.sleep(0.001)

        _verify_cpu_metering(1 * (10 ** 6))
        _verify_cpu_metering(3 * (10 ** 6))
        _verify_cpu_metering(2 * (10 ** 6))

    # the following apply to all instance resource pollsters but are tested
    # here alone.

    def test_get_metadata(self):
        mgr = manager.AgentManager(0, self.CONF)
        pollster = instance_stats.CPUPollster(self.CONF)
        samples = list(pollster.get_samples(mgr, {}, [self.instance]))
        self.assertEqual(1, len(samples))
        self.assertEqual(1, samples[0].resource_metadata['vcpus'])
        self.assertEqual(512, samples[0].resource_metadata['memory_mb'])
        self.assertEqual(20, samples[0].resource_metadata['disk_gb'])
        self.assertEqual(20, samples[0].resource_metadata['root_gb'])
        self.assertEqual(0, samples[0].resource_metadata['ephemeral_gb'])
        self.assertEqual('active', samples[0].resource_metadata['status'])
        self.assertEqual('active', samples[0].resource_metadata['state'])
        self.assertIsNone(samples[0].resource_metadata['task_state'])
        self.assertEqual(self.instance.flavor,
                         samples[0].resource_metadata['flavor'])
        self.assertEqual(self.instance.image['id'],
                         samples[0].resource_metadata['image_ref'])
        self.assertEqual(self.instance.image_meta,
                         samples[0].resource_metadata['image_meta'])

    def test_get_reserved_metadata_with_keys(self):
        self.CONF.set_override('reserved_metadata_keys', ['fqdn'])

        mgr = manager.AgentManager(0, self.CONF)
        pollster = instance_stats.CPUPollster(self.CONF)
        samples = list(pollster.get_samples(mgr, {}, [self.instance]))
        self.assertEqual({'fqdn': 'vm_fqdn',
                          'stack': '2cadc4b4-8789-123c-b4eg-edd2f0a9c128'},
                         samples[0].resource_metadata['user_metadata'])

    def test_get_reserved_metadata_with_namespace(self):
        mgr = manager.AgentManager(0, self.CONF)
        pollster = instance_stats.CPUPollster(self.CONF)
        samples = list(pollster.get_samples(mgr, {}, [self.instance]))
        self.assertEqual({'stack': '2cadc4b4-8789-123c-b4eg-edd2f0a9c128'},
                         samples[0].resource_metadata['user_metadata'])

        self.CONF.set_override('reserved_metadata_namespace', [])
        mgr = manager.AgentManager(0, self.CONF)
        pollster = instance_stats.CPUPollster(self.CONF)
        samples = list(pollster.get_samples(mgr, {}, [self.instance]))
        self.assertNotIn('user_metadata', samples[0].resource_metadata)

    def test_get_flavor_name_as_metadata_instance_type(self):
        mgr = manager.AgentManager(0, self.CONF)
        pollster = instance_stats.CPUPollster(self.CONF)
        samples = list(pollster.get_samples(mgr, {}, [self.instance]))
        self.assertEqual(1, len(samples))
        self.assertEqual('m1.small',
                         samples[0].resource_metadata['instance_type'])


class TestVCPUsPollster(base.TestPollsterBase):

    def test_get_samples(self):
        self._mock_inspect_instance(
            virt_inspector.InstanceStats(cpu_time=1 * (10 ** 6), cpu_number=1),
            virt_inspector.InstanceStats(cpu_time=3 * (10 ** 6), cpu_number=1),
            # cpu_time resets on instance restart
            virt_inspector.InstanceStats(cpu_time=2 * (10 ** 6), cpu_number=2),
        )

        mgr = manager.AgentManager(0, self.CONF)
        pollster = instance_stats.VCPUsPollster(self.CONF)

        def _verify_cpu_metering(expected_cpu_number):
            cache = {}
            samples = list(pollster.get_samples(mgr, cache, [self.instance]))
            self.assertEqual(1, len(samples))
            self.assertEqual({'vcpus'}, {s.name for s in samples})
            self.assertEqual(expected_cpu_number, samples[0].volume)
            # ensure elapsed time between polling cycles is non-zero
            time.sleep(0.001)

        _verify_cpu_metering(1)
        _verify_cpu_metering(1)
        _verify_cpu_metering(2)



#
# Copyright 2012 New Dream Network, LLC (DreamHost)
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
"""Base class for plugins.
"""

import abc

from stevedore import extension


class ExtensionLoadError(Exception):
    """Error of loading pollster plugin.

    PollsterBase provides a hook, setup_environment, called in pollster loading
    to setup required HW/SW dependency. Any exception from it would be
    propagated as ExtensionLoadError, then skip loading this pollster.
    """

    def __init__(self, msg=None):
        self.msg = msg


class PollsterPermanentError(Exception):
    """Permanent error when polling.

    When unrecoverable error happened in polling, pollster can raise this
    exception with failed resource to prevent itself from polling any more.
    Resource is one of parameter resources from get_samples that cause polling
    error.
    """

    def __init__(self, resources):
        self.fail_res_list = resources


class PollsterBase(metaclass=abc.ABCMeta):
    """Base class for plugins that support the polling API."""

    def setup_environment(self):
        """Setup required environment for pollster.

        Each subclass could overwrite it for specific usage. Any exception
        raised in this function would prevent pollster being loaded.
        """
        pass

    def __init__(self, conf):
        super().__init__()
        self.conf = conf
        try:
            self.setup_environment()
        except Exception as err:
            raise ExtensionLoadError(err)

    @property
    @abc.abstractmethod
    def default_discovery(self):
        """Default discovery to use for this pollster.

        There are three ways a pollster can get a list of resources to poll,
        listed here in ascending order of precedence:
        1. from the per-agent discovery,
        2. from the per-pollster discovery (defined here)
        3. from the per-pipeline configured discovery and/or per-pipeline
        configured static resources.

        If a pollster should only get resources from #1 or #3, this property
        should be set to None.
        """

    @abc.abstractmethod
    def get_samples(self, manager, cache, resources):
        """Return a sequence of Counter instances from polling the resources.

        :param manager: The service manager class invoking the plugin.
        :param cache: A dictionary to allow pollsters to pass data
                      between themselves when recomputing it would be
                      expensive (e.g., asking another service for a
                      list of objects).
        :param resources: A list of resources the pollster will get data
                          from. It's up to the specific pollster to decide
                          how to use it. It is usually supplied by a discovery,
                          see ``default_discovery`` for more information.

        """

    @classmethod
    def build_pollsters(cls, conf):
        """Return a list of tuple (name, pollster).

        The name is the meter name which the pollster would return, the
        pollster is a pollster object instance. The pollster which implements
        this method should be registered in the namespace of
        ceilometer.builder.xxx instead of ceilometer.poll.xxx.
        """
        return []

    @classmethod
    def get_pollsters_extensions(cls, conf):
        """Return a list of stevedore extensions.

        The returned stevedore extensions wrap the pollster object instances
        returned by build_pollsters.
        """
        extensions = []
        try:
            for name, pollster in cls.build_pollsters(conf):
                ext = extension.Extension(name, None, cls, pollster)
                extensions.append(ext)
        except Exception as err:
            raise ExtensionLoadError(err)
        return extensions


class DiscoveryBase(metaclass=abc.ABCMeta):
    KEYSTONE_REQUIRED_FOR_SERVICE = None
    """Service type required in keystone catalog to works"""

    def __init__(self, conf):
        self.conf = conf

    @abc.abstractmethod
    def discover(self, manager, param=None):
        """Discover resources to monitor.

        The most fine-grained discovery should be preferred, so the work is
        the most evenly distributed among multiple agents (if they exist).

        For example:
        if the pollster can separately poll individual resources, it should
        have its own discovery implementation to discover those resources. If
        it can only poll per-tenant, then the `TenantDiscovery` should be
        used. If even that is not possible, use `EndpointDiscovery` (see
        their respective docstrings).

        :param manager: The service manager class invoking the plugin.
        :param param: an optional parameter to guide the discovery
        """

    @property
    def group_id(self):
        """Return group id of this discovery.

        All running discoveries with the same group_id should return the same
        set of resources at a given point in time. By default, a discovery is
        put into a global group, meaning that all discoveries of its type
        running anywhere in the cloud, return the same set of resources.

        This property can be overridden to provide correct grouping of
        localized discoveries. For example, compute discovery is localized
        to a host, which is reflected in its group_id.

        A None value signifies that this discovery does not want to be part
        of workload partitioning at all.
        """
        return 'global'



#
# Copyright 2012 New Dream Network, LLC (DreamHost)
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
"""Tests for ceilometer/publisher/utils.py
"""
import json

from oslotest import base

from ceilometer.publisher import utils


class TestSignature(base.BaseTestCase):
    def test_compute_signature_change_key(self):
        sig1 = utils.compute_signature({'a': 'A', 'b': 'B'},
                                       'not-so-secret')
        sig2 = utils.compute_signature({'A': 'A', 'b': 'B'},
                                       'not-so-secret')
        self.assertNotEqual(sig1, sig2)

    def test_compute_signature_change_value(self):
        sig1 = utils.compute_signature({'a': 'A', 'b': 'B'},
                                       'not-so-secret')
        sig2 = utils.compute_signature({'a': 'a', 'b': 'B'},
                                       'not-so-secret')
        self.assertNotEqual(sig1, sig2)

    def test_compute_signature_same(self):
        sig1 = utils.compute_signature({'a': 'A', 'b': 'B'},
                                       'not-so-secret')
        sig2 = utils.compute_signature({'a': 'A', 'b': 'B'},
                                       'not-so-secret')
        self.assertEqual(sig1, sig2)

    def test_compute_signature_signed(self):
        data = {'a': 'A', 'b': 'B'}
        sig1 = utils.compute_signature(data, 'not-so-secret')
        data['message_signature'] = sig1
        sig2 = utils.compute_signature(data, 'not-so-secret')
        self.assertEqual(sig1, sig2)

    def test_compute_signature_use_configured_secret(self):
        data = {'a': 'A', 'b': 'B'}
        sig1 = utils.compute_signature(data, 'not-so-secret')
        sig2 = utils.compute_signature(data, 'different-value')
        self.assertNotEqual(sig1, sig2)

    def test_verify_signature_signed(self):
        data = {'a': 'A', 'b': 'B'}
        sig1 = utils.compute_signature(data, 'not-so-secret')
        data['message_signature'] = sig1
        self.assertTrue(utils.verify_signature(data, 'not-so-secret'))

    def test_verify_signature_unsigned(self):
        data = {'a': 'A', 'b': 'B'}
        self.assertFalse(utils.verify_signature(data, 'not-so-secret'))

    def test_verify_signature_incorrect(self):
        data = {'a': 'A', 'b': 'B',
                'message_signature': 'Not the same'}
        self.assertFalse(utils.verify_signature(data, 'not-so-secret'))

    def test_verify_signature_invalid_encoding(self):
        data = {'a': 'A', 'b': 'B',
                'message_signature': ''}
        self.assertFalse(utils.verify_signature(data, 'not-so-secret'))

    def test_verify_signature_unicode(self):
        data = {'a': 'A', 'b': 'B',
                'message_signature': ''}
        self.assertFalse(utils.verify_signature(data, 'not-so-secret'))

    def test_verify_signature_nested(self):
        data = {'a': 'A',
                'b': 'B',
                'nested': {'a': 'A',
                           'b': 'B',
                           },
                }
        data['message_signature'] = utils.compute_signature(
            data,
            'not-so-secret')
        self.assertTrue(utils.verify_signature(data, 'not-so-secret'))

    def test_verify_signature_nested_json(self):
        data = {'a': 'A',
                'b': 'B',
                'nested': {'a': 'A',
                           'b': 'B',
                           'c': ('c',),
                           'd': ['d']
                           },
                }
        data['message_signature'] = utils.compute_signature(
            data,
            'not-so-secret')
        jsondata = json.loads(json.dumps(data))
        self.assertTrue(utils.verify_signature(jsondata, 'not-so-secret'))

    def test_verify_unicode_symbols(self):
        data = {'a\xe9\u0437': 'A',
                'b': 'B\xe9\u0437'
                }
        data['message_signature'] = utils.compute_signature(
            data,
            'not-so-secret')
        jsondata = json.loads(json.dumps(data))
        self.assertTrue(utils.verify_signature(jsondata, 'not-so-secret'))

    def test_verify_no_secret(self):
        data = {'a': 'A', 'b': 'B'}
        self.assertTrue(utils.verify_signature(data, ''))


class TestUtils(base.BaseTestCase):
    def test_recursive_keypairs(self):
        data = {'a': 'A', 'b': 'B',
                'nested': {'a': 'A', 'b': 'B'}}
        pairs = list(utils.recursive_keypairs(data))
        self.assertEqual([('a', 'A'), ('b', 'B'),
                          ('nested:a', 'A'), ('nested:b', 'B')],
                         pairs)

    def test_recursive_keypairs_with_separator(self):
        data = {'a': 'A',
                'b': 'B',
                'nested': {'a': 'A',
                           'b': 'B',
                           },
                }
        separator = '.'
        pairs = list(utils.recursive_keypairs(data, separator))
        self.assertEqual([('a', 'A'),
                          ('b', 'B'),
                          ('nested.a', 'A'),
                          ('nested.b', 'B')],
                         pairs)

    def test_recursive_keypairs_with_list_of_dict(self):
        small = 1
        big = 1 << 64
        expected = [('a', 'A'),
                    ('b', 'B'),
                    ('nested:list', [{small: 99, big: 42}])]
        data = {'a': 'A',
                'b': 'B',
                'nested': {'list': [{small: 99, big: 42}]}}
        pairs = list(utils.recursive_keypairs(data))
        self.assertEqual(len(expected), len(pairs))
        for k, v in pairs:
            # the keys 1 and 1<<64 cause a hash collision on 64bit platforms
            if k == 'nested:list':
                self.assertIn(v,
                              [[{small: 99, big: 42}],
                               [{big: 42, small: 99}]])
            else:
                self.assertIn((k, v), expected)



#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import openstack
from oslo_config import cfg

from ceilometer import keystone_client

SERVICE_OPTS = [
    cfg.StrOpt('manila',
               default='sharev2',
               help='Manila service type.'),
]


class Client:
    """A client which gets information via openstacksdk."""

    def __init__(self, conf):
        """Initialize a Manila client object."""
        creds = conf.service_credentials

        self.conn = openstack.connection.Connection(
            session=keystone_client.get_session(conf),
            region_name=creds.region_name,
            shared_file_system_interface=creds.interface,
            shared_file_system_service_type=conf.service_types.manila
        )

    def shares_list(self):
        """Return a list of shares from all projects."""
        return self.conn.shared_file_system.shares(all_projects=True)



#
# Copyright 2024 Juan Larriba
# Copyright 2024 Red Hat, Inc
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import prometheus_client as prom

CEILOMETER_REGISTRY = prom.CollectorRegistry()


def export(prom_iface, prom_port, tls_cert=None, tls_key=None):
    prom.start_http_server(port=prom_port,
                           addr=prom_iface,
                           registry=CEILOMETER_REGISTRY,
                           certfile=tls_cert,
                           keyfile=tls_key)


def collect_metrics(samples):
    for sample in samples:
        name = "ceilometer_" + sample['counter_name'].replace('.', '_')
        labels = _gen_labels(sample)

        metric = CEILOMETER_REGISTRY._names_to_collectors.get(name, None)

        if metric is None:
            metric = prom.Gauge(name=name, documentation="",
                                labelnames=labels['keys'],
                                registry=CEILOMETER_REGISTRY)
        metric.labels(*labels['values']).set(sample['counter_volume'])


def purge_stale_metrics(pollster):
    metric_cleared = False

    metric_name = "ceilometer_" + pollster.replace('.', '_')
    metric = CEILOMETER_REGISTRY._names_to_collectors.get(metric_name, None)
    if not metric_cleared:
        if metric:
            CEILOMETER_REGISTRY.unregister(metric)
            metric = None
        metric_cleared = True


def _gen_labels(sample):
    labels = dict(keys=[], values=[])
    cNameShards = sample['counter_name'].split(".")
    ctype = ''

    plugin = cNameShards[0]
    pluginVal = sample['resource_id']
    if len(cNameShards) > 2:
        pluginVal = cNameShards[2]

    if len(cNameShards) > 1:
        ctype = cNameShards[1]
    else:
        ctype = cNameShards[0]

    labels['keys'].append(plugin)
    labels['values'].append(pluginVal)

    labels['keys'].append("publisher")
    labels['values'].append("ceilometer")

    labels['keys'].append("type")
    labels['values'].append(ctype)

    if sample.get('counter_name'):
        labels['keys'].append("counter")
        labels['values'].append(sample['counter_name'])

    if sample.get('project_id'):
        labels['keys'].append("project")
        labels['values'].append(sample['project_id'])

    if sample.get('project_name'):
        labels['keys'].append("project_name")
        labels['values'].append(sample['project_name'])

    if sample.get('user_id'):
        labels['keys'].append("user")
        labels['values'].append(sample['user_id'])

    if sample.get('user_name'):
        labels['keys'].append("user_name")
        labels['values'].append(sample['user_name'])

    if sample.get('counter_unit'):
        labels['keys'].append("unit")
        labels['values'].append(sample['counter_unit'])

    if sample.get('resource_id'):
        labels['keys'].append("resource")
        labels['values'].append(sample['resource_id'])

    if sample.get('resource_metadata'):
        resource_metadata = sample['resource_metadata']

        if resource_metadata.get('host'):
            labels['keys'].append("vm_instance")
            labels['values'].append(resource_metadata['host'])

        if resource_metadata.get('display_name'):
            value = resource_metadata['display_name']
            if resource_metadata.get('name'):
                value = ':'.join([value, resource_metadata['name']])
            labels['keys'].append("resource_name")
            labels['values'].append(value)
        elif resource_metadata.get('name'):
            labels['keys'].append("resource_name")
            labels['values'].append(resource_metadata['name'])

        # NOTE(jwysogla): The prometheus_client library doesn't support
        # variable count of labels for the same metric. That's why the
        # prometheus exporter cannot support custom metric labels added
        # with the --property metering.<label_key>=<label_value> when
        # creating a server. This still works with publishers though.
        # The "server_group" label is used for autoscaling and so it's
        # the only one getting parsed. To always have the same number
        # of labels, it's added to all metrics and where there isn't a
        # value defined, it's substituted with "none".
        user_metadata = resource_metadata.get('user_metadata', {})
        if user_metadata.get('server_group'):
            labels['keys'].append('server_group')
            labels['values'].append(user_metadata['server_group'])
        else:
            labels['keys'].append('server_group')
            labels['values'].append('none')
        if resource_metadata.get('alarm_state', '') != '':
            labels['keys'].append('state')
            labels['values'].append(resource_metadata['alarm_state'])

        # Add availability_zone for loadbalancer metrics
        if sample.get('counter_name', '').startswith('loadbalancer'):
            labels['keys'].append('availability_zone')
            az = resource_metadata.get('availability_zone')
            labels['values'].append(az if az else '')

        if resource_metadata.get('flavor'):
            flavor = resource_metadata.get('flavor')
            if flavor.get('id'):
                labels['keys'].append("flavor_id")
                labels['values'].append(flavor['id'])
            if flavor.get('name'):
                labels['keys'].append("flavor_name")
                labels['values'].append(flavor['name'])

    return labels



#
# Copyright 2014 Red Hat Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
"""Tests for ceilometer/central/manager.py"""

from unittest import mock

from oslotest import base

from ceilometer.polling.discovery import endpoint
from ceilometer.polling.discovery import localnode
from ceilometer.polling.discovery import tenant as project
from ceilometer import service


class TestEndpointDiscovery(base.BaseTestCase):

    def setUp(self):
        super().setUp()
        CONF = service.prepare_service([], [])
        CONF.set_override('interface', 'publicURL',
                          group='service_credentials')
        CONF.set_override('region_name', 'test-region-name',
                          group='service_credentials')
        self.discovery = endpoint.EndpointDiscovery(CONF)
        self.manager = mock.MagicMock()
        self.catalog = (self.manager.keystone.session.auth.get_access.
                        return_value.service_catalog)

    def test_keystone_called(self):
        self.discovery.discover(self.manager, param='test-service-type')
        expected = [mock.call(service_type='test-service-type',
                              service_name=None,
                              interface='publicURL',
                              region_name='test-region-name')]
        self.assertEqual(expected, self.catalog.get_urls.call_args_list)

    def test_keystone_called_no_service_type(self):
        self.discovery.discover(self.manager)
        expected = [mock.call(service_type=None,
                              service_name=None,
                              interface='publicURL',
                              region_name='test-region-name')]
        self.assertEqual(expected,
                         self.catalog.get_urls
                         .call_args_list)

    def test_keystone_called_no_endpoints(self):
        self.catalog.get_urls.return_value = []
        self.assertEqual([], self.discovery.discover(self.manager))


class TestLocalnodeDiscovery(base.BaseTestCase):
    def setUp(self):
        super().setUp()
        self.conf = service.prepare_service([], [])
        self.discovery = localnode.LocalNodeDiscovery(self.conf)
        self.manager = mock.MagicMock()

    def test_lockalnode_discovery(self):
        self.assertEqual([self.conf.host],
                         self.discovery.discover(self.manager))


class TestProjectDiscovery(base.BaseTestCase):
    def prepare_mock_data(self):
        domain_heat = mock.MagicMock()
        domain_heat.id = '2f42ab40b7ad4140815ef830d816a16c'
        domain_heat.name = 'heat'
        domain_heat.enabled = True
        domain_heat.links = {
            'self': 'http://192.168.1.1/identity/v3/domains/'
                    '2f42ab40b7ad4140815ef830d816a16c'}

        domain_default = mock.MagicMock()
        domain_default.id = 'default'
        domain_default.name = 'Default'
        domain_default.enabled = True
        domain_default.links = {
            'self': 'http://192.168.1.1/identity/v3/domains/default'}

        project_admin = mock.MagicMock()
        project_admin.id = '2ce92449a23145ef9c539f3327960ce3'
        project_admin.name = 'admin'
        project_admin.parent_id = 'default'
        project_admin.domain_id = 'default'
        project_admin.is_domain = False
        project_admin.enabled = True
        project_admin.links = {
            'self': 'http://192.168.4.46/identity/v3/projects/'
                    '2ce92449a23145ef9c539f3327960ce3'},

        project_service = mock.MagicMock()
        project_service.id = '9bf93b86bca04e3b815f86a5de083adc'
        project_service.name = 'service'
        project_service.parent_id = 'default'
        project_service.domain_id = 'default'
        project_service.is_domain = False
        project_service.enabled = True
        project_service.links = {
            'self': 'http://192.168.4.46/identity/v3/projects/'
                    '9bf93b86bca04e3b815f86a5de083adc'}

        project_demo = mock.MagicMock()
        project_demo.id = '57d96b9af18d43bb9d047f436279b0be'
        project_demo.name = 'demo'
        project_demo.parent_id = 'default'
        project_demo.domain_id = 'default'
        project_demo.is_domain = False
        project_demo.enabled = True
        project_demo.links = {
            'self': 'http://192.168.4.46/identity/v3/projects/'
                    '57d96b9af18d43bb9d047f436279b0be'}

        self.domains = [domain_heat, domain_default]
        self.default_domain_projects = [project_admin, project_service]
        self.heat_domain_projects = [project_demo]

    def side_effect(self, domain=None, **kwargs):
        if not domain or domain.name == 'Default':
            return self.default_domain_projects
        elif domain.name == 'heat':
            return self.heat_domain_projects
        else:
            return []

    def setUp(self):
        super().setUp()
        self.CONF = service.prepare_service([], [])
        self.discovery = project.TenantDiscovery(self.CONF)
        self.prepare_mock_data()
        self.manager = mock.MagicMock()
        self.manager.keystone.projects.list.side_effect = self.side_effect

    def test_project_discovery(self):
        self.manager.keystone.domains.list.return_value = self.domains
        result = self.discovery.discover(self.manager)
        self.assertEqual(len(result), 3)
        self.manager.keystone.domains.list.assert_called_once_with()
        self.assertEqual(
            self.manager.keystone.projects.list.mock_calls,
            [mock.call(domain) for domain in self.domains])

    def test_project_discovery_ignore_disabled_projects(self):
        self.CONF.set_override("ignore_disabled_projects",
                               True, group="polling")
        self.manager.keystone.domains.list.return_value = self.domains
        result = self.discovery.discover(self.manager)
        self.assertEqual(len(result), 3)
        self.manager.keystone.domains.list.assert_called_once_with(
            enabled=True)
        self.assertEqual(
            self.manager.keystone.projects.list.mock_calls,
            [mock.call(domain, enabled=True) for domain in self.domains])



#
# Copyright 2012 New Dream Network, LLC (DreamHost)
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
"""Common code for working with images
"""

from ceilometer.polling import plugin_base
from ceilometer import sample


class _Base(plugin_base.PollsterBase):
    @property
    def default_discovery(self):
        return 'images'

    @staticmethod
    def extract_image_metadata(image):
        return {
            k: getattr(image, k)
            for k in [
                "status",
                "visibility",
                "name",
                "container_format",
                "created_at",
                "disk_format",
                "updated_at",
                "min_disk",
                "is_protected",
                "checksum",
                "min_ram",
                "tags",
                "virtual_size"
            ]
        }


class ImageSizePollster(_Base):
    def get_samples(self, manager, cache, resources):
        for image in resources:
            yield sample.Sample(
                name='image.size',
                type=sample.TYPE_GAUGE,
                unit='B',
                volume=image.size,
                user_id=None,
                project_id=image.owner,
                resource_id=image.id,
                resource_metadata=self.extract_image_metadata(image),
            )



# Copyright 2014 Red Hat, Inc
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_log import log

from ceilometer.polling import plugin_base as plugin

LOG = log.getLogger(__name__)


class TenantDiscovery(plugin.DiscoveryBase):
    """Discovery that supplies keystone tenants.

    This discovery should be used when the pollster's work can't be divided
    into smaller pieces than per-tenants. Example of this is the Swift
    pollster, which polls account details and does so per-project.
    """

    def discover(self, manager, param=None):
        filters = {}
        # When ignore_disabled_projects is enabled, add a filter
        # to the Keystone API queries so that only enabled projects,
        # and projects in enabled domains, are returned.
        if self.conf.polling.ignore_disabled_projects:
            filters["enabled"] = True

        domains = manager.keystone.domains.list(**filters)
        LOG.debug(
            "Found %i %sKeystone domains",
            len(domains),
            "enabled " if self.conf.polling.ignore_disabled_projects else "")

        tenants = []
        projects_log_message = (
            "Found %i enabled projects in domain %s"
            if self.conf.polling.ignore_disabled_projects
            else "Found %i projects in domain %s")
        for domain in domains:
            domain_projects = manager.keystone.projects.list(
                domain,
                **filters)
            LOG.debug(projects_log_message, len(domain_projects), domain.name)
            tenants.extend(domain_projects)

        return tenants



#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from urllib import parse as urlparse

from ceilometer import keystone_client
from ceilometer import publisher

from zaqarclient.queues.v2 import client as zaqarclient

DEFAULT_TTL = 3600


class ZaqarPublisher(publisher.ConfigPublisherBase):
    """Publish metering data to a Zaqar queue.

    The target queue name must be configured in the ceilometer pipeline
    configuration file. The TTL can also optionally be specified as a query
    argument::

        meter:
            - name: meter_zaqar
            meters:
                - "*"
            sinks:
                - zaqar_sink
        sinks:
            - name: zaqar_sink
            publishers:
                - zaqar://?queue=meter_queue&ttl=1200

    The credentials to access Zaqar must be set in the [zaqar] section in the
    configuration.
    """

    def __init__(self, conf, parsed_url):
        super().__init__(conf, parsed_url)
        options = urlparse.parse_qs(parsed_url.query)
        self.queue_name = options.get('queue', [None])[0]
        if not self.queue_name:
            raise ValueError('Must specify a queue in the zaqar publisher')
        self.ttl = int(options.pop('ttl', [DEFAULT_TTL])[0])
        self._client = None

    @property
    def client(self):
        if self._client is None:
            session = keystone_client.get_session(
                self.conf, group=self.conf.zaqar.auth_section)
            self._client = zaqarclient.Client(session=session)
        return self._client

    def publish_samples(self, samples):
        """Send a metering message for publishing

        :param samples: Samples from pipeline.
        """
        queue = self.client.queue(self.queue_name)
        messages = [{'body': sample.as_dict(), 'ttl': self.ttl}
                    for sample in samples]
        queue.post(messages)

    def publish_events(self, events):
        """Send an event message for publishing

        :param events: events from pipeline.
        """
        queue = self.client.queue(self.queue_name)
        messages = [{'body': event.serialize(), 'ttl': self.ttl}
                    for event in events]
        queue.post(messages)



#
# Copyright 2014 Red Hat, Inc
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import hashlib
from lxml import etree
import operator
import threading

import cachetools
from novaclient import exceptions
from oslo_config import cfg
from oslo_log import log
from oslo_utils import timeutils

from ceilometer.compute.virt.libvirt import utils as libvirt_utils
from ceilometer import nova_client
from ceilometer.polling import plugin_base

OPTS = [
    cfg.StrOpt('instance_discovery_method',
               default='libvirt_metadata',
               choices=[('naive', 'poll nova to get all instances'),
                        ('workload_partitioning',
                         'poll nova to get instances of the compute'),
                        ('libvirt_metadata',
                         'get instances from libvirt metadata but without '
                         'instance metadata (recommended)')],
               help="Ceilometer offers many methods to discover the instance "
                    "running on a compute node"),
    cfg.IntOpt('resource_update_interval',
               default=0,
               min=0,
               help="New instances will be discovered periodically based"
                    " on this option (in seconds). By default, "
                    "the agent discovers instances according to pipeline "
                    "polling interval. If option is greater than 0, "
                    "the instance list to poll will be updated based "
                    "on this option's interval. Measurements relating "
                    "to the instances will match intervals "
                    "defined in pipeline. This option is only used "
                    "for agent polling to Nova API, so it will work only "
                    "when 'instance_discovery_method' is set to 'naive'."),
    cfg.IntOpt('resource_cache_expiry',
               default=3600,
               min=0,
               help="The expiry to totally refresh the instances resource "
                    "cache, since the instance may be migrated to another "
                    "host, we need to clean the legacy instances info in "
                    "local cache by totally refreshing the local cache. "
                    "The minimum should be the value of the config option "
                    "of resource_update_interval. This option is only used "
                    "for agent polling to Nova API, so it will work only "
                    "when 'instance_discovery_method' is set to 'naive'."),
    cfg.BoolOpt('fetch_extra_metadata',
                default=True,
                help="Whether or not additional instance attributes that "
                     "require Nova API queries should be fetched. Currently "
                     "the only value that requires fetching from Nova API is "
                     "'metadata', the attribute storing user-configured "
                     "server metadata, which is used to fill out some "
                     "optional fields such as the server group of an "
                     "instance. fetch_extra_metadata is currently set to "
                     "True by default, but to reduce the load on Nova API "
                     "this will be changed to False in a future release."),
]

LOG = log.getLogger(__name__)


class NovaLikeServer:
    def __init__(self, **kwargs):
        self.id = kwargs.pop('id')
        for k, v in kwargs.items():
            setattr(self, k, v)

    def __repr__(self):
        return '<NovaLikeServer: %s>' % getattr(self, 'name', 'unknown-name')

    def __eq__(self, other):
        return self.id == other.id


class InstanceDiscovery(plugin_base.DiscoveryBase):
    method = None

    def __init__(self, conf):
        super().__init__(conf)
        if not self.method:
            self.method = conf.compute.instance_discovery_method

        self.nova_cli = nova_client.Client(conf)
        self.expiration_time = conf.compute.resource_update_interval
        self.cache_expiry = conf.compute.resource_cache_expiry
        if self.method == "libvirt_metadata":
            # 4096 resources on a compute should be enough :)
            self._flavor_id_cache = cachetools.LRUCache(4096)
            self._server_cache = cachetools.LRUCache(4096)
        else:
            self.lock = threading.Lock()
            self.instances = {}
            self.last_run = None
            self.last_cache_expire = None

    @property
    def connection(self):
        return libvirt_utils.refresh_libvirt_connection(self.conf, self)

    def discover(self, manager, param=None):
        """Discover resources to monitor."""
        if self.method != "libvirt_metadata":
            return self.discover_nova_polling(manager, param=None)
        else:
            return self.discover_libvirt_polling(manager, param=None)

    @staticmethod
    def _safe_find_int(xml, path):
        elem = xml.find("./%s" % path)
        if elem is not None:
            return int(elem.text)
        return 0

    def _get_flavor_id(self, flavor_xml, instance_id):
        flavor_name = flavor_xml.attrib["name"]
        # Flavor ID is available in libvirt metadata from 2025.2 onwards.
        flavor_id = flavor_xml.attrib.get("id")
        if flavor_id:
            return flavor_id
        # If not found in libvirt metadata, fallback to API queries.
        # If we already have the server metadata get the flavor ID from there.
        if self.conf.compute.fetch_extra_metadata:
            server = self.get_server(instance_id)
            if server:
                return server.flavor["id"]
        # If server metadata is not otherwise fetched, or the query failed,
        # query just the flavor for better cache hit rates.
        return (self.get_flavor_id(flavor_name) or flavor_name)

    def _get_flavor_extra_specs(self, flavor_xml):
        # Extra specs are available in libvirt metadata from 2025.2 onwards.
        # Note that this checks for existence of the element, not whether
        # or not it is empty, as it *can* exist but have nothing set in it.
        extra_specs = flavor_xml.find("./extraSpecs")
        if extra_specs is not None:
            return {
                extra_spec.attrib["name"]: extra_spec.text
                for extra_spec in extra_specs.findall("./extraSpec")}
        # If not found in libvirt metadata, return None to signify
        # "not fetched", as we don't support performing additional
        # API queries just for the extra specs.
        return None

    @cachetools.cachedmethod(operator.attrgetter('_flavor_id_cache'))
    def get_flavor_id(self, name):
        LOG.debug("Querying metadata for flavor %s from Nova API", name)
        try:
            return self.nova_cli.nova_client.flavors.find(
                name=name,
                is_public=None).id
        except exceptions.NotFound:
            return None

    @cachetools.cachedmethod(operator.attrgetter('_server_cache'))
    def get_server(self, uuid):
        LOG.debug("Querying metadata for instance %s from Nova API", uuid)
        try:
            return self.nova_cli.nova_client.servers.get(uuid)
        except exceptions.NotFound:
            return None

    @libvirt_utils.retry_on_disconnect
    def discover_libvirt_polling(self, manager, param=None):
        instances = []
        for domain in self.connection.listAllDomains():
            instance_id = domain.UUIDString()
            xml_string = libvirt_utils.instance_metadata(domain)
            if xml_string is None:
                continue

            full_xml = etree.fromstring(domain.XMLDesc())
            os_type_xml = full_xml.find("./os/type")
            metadata_xml = etree.fromstring(xml_string)

            try:
                flavor_xml = metadata_xml.find(
                    "./flavor")
                user_id = metadata_xml.find(
                    "./owner/user").attrib["uuid"]
                project_id = metadata_xml.find(
                    "./owner/project").attrib["uuid"]
                instance_name = metadata_xml.find(
                    "./name").text
                instance_arch = os_type_xml.attrib["arch"]

                extra_specs = self._get_flavor_extra_specs(flavor_xml)

                flavor = {
                    "id": self._get_flavor_id(flavor_xml, instance_id),
                    "name": flavor_xml.attrib["name"],
                    "vcpus": self._safe_find_int(flavor_xml, "vcpus"),
                    "ram": self._safe_find_int(flavor_xml, "memory"),
                    "disk": self._safe_find_int(flavor_xml, "disk"),
                    "ephemeral": self._safe_find_int(flavor_xml, "ephemeral"),
                    "swap": self._safe_find_int(flavor_xml, "swap"),
                }
                if extra_specs is not None:
                    flavor["extra_specs"] = extra_specs

                image_xml = metadata_xml.find("./root[@type='image']")
                image = ({'id': image_xml.attrib['uuid']}
                         if image_xml is not None else None)

                image_meta_xml = metadata_xml.find("./image")
                if image_meta_xml is not None:
                    # If the <image> element exists at all, Nova supports
                    # image_meta in libvirt metadata. Add it to the instance
                    # attributes even if all the required values are empty.
                    image_meta = {}
                    base_image_ref = image_meta_xml.attrib.get("uuid")
                    if base_image_ref is not None:
                        image_meta["base_image_ref"] = base_image_ref
                    # The following properties get special treatment
                    # because they are set as such in SM_INHERITABLE_KEYS,
                    # as defined in nova/utils.py.
                    container_format_xml = image_meta_xml.find(
                        "./containerFormat")
                    if container_format_xml is not None:
                        image_meta["container_format"] = (
                            container_format_xml.text)
                    disk_format_xml = image_meta_xml.find("./diskFormat")
                    if disk_format_xml is not None:
                        image_meta["disk_format"] = disk_format_xml.text
                    min_disk_xml = image_meta_xml.find("./minDisk")
                    if min_disk_xml is not None:
                        image_meta["min_disk"] = min_disk_xml.text
                    min_ram_xml = image_meta_xml.find("./minRam")
                    if min_ram_xml is not None:
                        image_meta["min_ram"] = min_ram_xml.text
                    # Get additional properties defined in image_meta.
                    properties_xml = image_meta_xml.find("./properties")
                    if properties_xml is not None:
                        for prop in properties_xml.findall("./property"):
                            image_meta[prop.attrib["name"]] = prop.text
                else:
                    # None for "no image_meta found".
                    image_meta = None

                # Getting the server metadata requires expensive Nova API
                # queries, and may potentially contain sensitive user info,
                # so it is only fetched when configured to do so.
                if self.conf.compute.fetch_extra_metadata:
                    server = self.get_server(instance_id)
                    metadata = server.metadata if server is not None else {}
                else:
                    metadata = {}
            except AttributeError:
                LOG.error(
                    "Fail to get domain uuid %s metadata: "
                    "metadata was missing expected attributes",
                    instance_id)
                continue

            dom_state = domain.state()[0]
            vm_state = libvirt_utils.LIBVIRT_POWER_STATE.get(dom_state)
            status = libvirt_utils.LIBVIRT_STATUS.get(dom_state)

            # From:
            # https://github.com/openstack/nova/blob/852f40fd0c6e9d8878212ff3120556668023f1c4/nova/api/openstack/compute/views/servers.py#L214-L220
            host_id = hashlib.sha224(
                (project_id + self.conf.host).encode('utf-8')).hexdigest()

            instance_data = {
                "id": instance_id,
                "name": instance_name,
                "flavor": flavor,
                "image": image,
                "os_type": os_type_xml.text,
                "architecture": instance_arch,

                "OS-EXT-SRV-ATTR:instance_name": domain.name(),
                "OS-EXT-SRV-ATTR:host": self.conf.host,
                "OS-EXT-STS:vm_state": vm_state,

                "tenant_id": project_id,
                "user_id": user_id,

                "hostId": host_id,
                "status": status,

                # NOTE(sileht): Other fields that Ceilometer tracks
                # where we can't get the value here, but their are
                # retrieved by notification
                "metadata": metadata,
                # "OS-EXT-STS:task_state"
                # 'reservation_id',
                # 'OS-EXT-AZ:availability_zone',
                # 'kernel_id',
                # 'ramdisk_id',
                # some image detail
            }
            if image_meta is not None:
                instance_data["image_meta"] = image_meta

            LOG.debug("instance data: %s", instance_data)
            instances.append(NovaLikeServer(**instance_data))
        return instances

    def discover_nova_polling(self, manager, param=None):
        secs_from_last_update = 0
        utc_now = timeutils.utcnow(True)
        secs_from_last_expire = 0
        if self.last_run:
            secs_from_last_update = timeutils.delta_seconds(
                self.last_run, utc_now)
        if self.last_cache_expire:
            secs_from_last_expire = timeutils.delta_seconds(
                self.last_cache_expire, utc_now)

        instances = []
        # NOTE(ityaptin) we update make a nova request only if
        # it's a first discovery or resources expired
        with self.lock:
            if (not self.last_run or secs_from_last_update >=
                    self.expiration_time):
                try:
                    if (secs_from_last_expire < self.cache_expiry and
                            self.last_run):
                        since = self.last_run.isoformat()
                    else:
                        since = None
                        self.instances.clear()
                        self.last_cache_expire = utc_now
                    instances = self.nova_cli.instance_get_all_by_host(
                        self.conf.host, since)
                    self.last_run = utc_now
                except Exception:
                    # NOTE(zqfan): instance_get_all_by_host is wrapped and will
                    # log exception when there is any error. It is no need to
                    #  raise it again and print one more time.
                    return []

            for instance in instances:
                if getattr(instance, 'OS-EXT-STS:vm_state', None) in [
                   'deleted', 'error']:
                    self.instances.pop(instance.id, None)
                else:
                    self.instances[instance.id] = instance

        return self.instances.values()

    @property
    def group_id(self):
        return self.conf.host



#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from ceilometer.polling import manager
from ceilometer import service
import ceilometer.tests.base as base
from ceilometer.volume import cinder

VOLUME_LIST = [
    type('Volume', (object,),
         {'migration_status': None,
          'attachments': [
              {'server_id': '1ae69721-d071-4156-a2bd-b11bb43ec2e3',
               'attachment_id': 'f903d95e-f999-4a34-8be7-119eadd9bb4f',
               'attached_at': '2016-07-14T03:55:57.000000',
               'host_name': None,
               'volume_id': 'd94c18fb-b680-4912-9741-da69ee83c94f',
               'device': '/dev/vdb',
               'id': 'd94c18fb-b680-4912-9741-da69ee83c94f'}],
          'links': [{
              'href': 'http://fake_link3',
              'rel': 'self'},
              {
                  'href': 'http://fake_link4',
                  'rel': 'bookmark'}],
          'availability_zone': 'nova',
          'os-vol-host-attr:host': 'test@lvmdriver-1#lvmdriver-1',
          'encrypted': False,
          'updated_at': '2016-07-14T03:55:57.000000',
          'replication_status': 'disabled',
          'snapshot_id': None,
          'id': 'd94c18fb-b680-4912-9741-da69ee83c94f',
          'size': 1,
          'user_id': 'be255bd31eb944578000fc762fde6dcf',
          'os-vol-tenant-attr:tenant_id': '6824974c08974d4db864bbaa6bc08303',
          'os-vol-mig-status-attr:migstat': None,
          'metadata': {'readonly': 'False', 'attached_mode': 'rw'},
          'status': 'in-use',
          'description': None,
          'multiattach': False,
          'source_volid': None,
          'consistencygroup_id': None,
          "volume_image_metadata": {
              "checksum": "17d9daa4fb8e20b0f6b7dec0d46fdddf",
              "container_format": "bare",
              "disk_format": "raw",
              "hw_disk_bus": "scsi",
              "hw_scsi_model": "virtio-scsi",
              "image_id": "f0019ee3-523c-45ab-b0b6-3adc529673e7",
              "image_name": "debian-jessie-scsi",
              "min_disk": "0",
              "min_ram": "0",
              "size": "1572864000"
          },
          'os-vol-mig-status-attr:name_id': None,
          'group_id': None,
          'provider_id': None,
          'shared_targets': False,
          'service_uuid': '2f6b5a18-0cd5-4421-b97e-d2c3e85ed758',
          'cluster_name': None,
          'volume_type_id': '65a9f65a-4696-4435-a09d-bc44d797c529',
          'name': None,
          'bootable': 'false',
          'created_at': '2016-06-23T08:27:45.000000',
          'volume_type': 'lvmdriver-1'})
]

SNAPSHOT_LIST = [
    type('VolumeSnapshot', (object,),
         {'status': 'available',
          'os-extended-snapshot-attributes:progress': '100%',
          'description': None,
          'os-extended-snapshot-attributes:project_id':
              '6824974c08974d4db864bbaa6bc08303',
          'size': 1,
          'user_id': 'be255bd31eb944578000fc762fde6dcf',
          'updated_at': '2016-10-19T07:56:55.000000',
          'id': 'b1ea6783-f952-491e-a4ed-23a6a562e1cf',
          'volume_id': '6f27bc42-c834-49ea-ae75-8d1073b37806',
          'metadata': {},
          'created_at': '2016-10-19T07:56:55.000000',
          "group_snapshot_id": None,
          'name': None})
]

BACKUP_LIST = [
    type('VolumeBackup', (object,),
         {'status': 'available',
          'object_count': 0,
          'container': None,
          'name': None,
          'links': [{
              'href': 'http://fake_urla',
              'rel': 'self'}, {
              'href': 'http://fake_urlb',
              'rel': 'bookmark'}],
          'availability_zone': 'nova',
          'created_at': '2016-10-19T06:55:23.000000',
          'snapshot_id': None,
          'updated_at': '2016-10-19T06:55:23.000000',
          'data_timestamp': '2016-10-19T06:55:23.000000',
          'description': None,
          'has_dependent_backups': False,
          'volume_id': '6f27bc42-c834-49ea-ae75-8d1073b37806',
          'os-backup-project-attr:project_id':
              '6824974c08974d4db864bbaa6bc08303',
          'fail_reason': "",
          'is_incremental': False,
          'metadata': {},
          'user_id': 'be255bd31eb944578000fc762fde6dcf',
          'id': '75a52125-85ff-4a8d-b2aa-580f3b22273f',
          'size': 1})
]

POOL_LIST = [
    type('VolumePool', (object,),
         {'name': 'localhost.localdomain@lvmdriver-1#lvmdriver-1',
          'pool_name': 'lvmdriver-1',
          'total_capacity_gb': 28.5,
          'free_capacity_gb': 28.39,
          'reserved_percentage': 0,
          'location_info':
              'LVMVolumeDriver:localhost.localdomain:stack-volumes:thin:0',
          'QoS_support': False,
          'provisioned_capacity_gb': 4.0,
          'max_over_subscription_ratio': 20.0,
          'thin_provisioning_support': True,
          'thick_provisioning_support': False,
          'total_volumes': 3,
          'filter_function': None,
          'goodness_function': None,
          'multiattach': True,
          'backend_state': 'up',
          'allocated_capacity_gb': 4,
          'cacheable': True,
          'volume_backend_name': 'lvmdriver-1',
          'storage_protocol': 'iSCSI',
          'vendor_name': 'Open Source',
          'driver_version': '3.0.0',
          'timestamp': '2025-03-21T14:19:02.901750'}),
    type('VolumePool', (object,),
         {'name': 'cinder-3ceee-volume-ceph-0@ceph#ceph',
          'vendor_name': 'Open Source',
          'driver_version': '1.3.0',
          'storage_protocol': 'ceph',
          'total_capacity_gb': 85.0,
          'free_capacity_gb': 85.0,
          'reserved_percentage': 0,
          'multiattach': True,
          'thin_provisioning_support': True,
          'max_over_subscription_ratio': '20.0',
          'location_info':
          'ceph:/etc/ceph/ceph.conf:a94b63c4e:openstack:volumes',
          'backend_state': 'up',
          'qos_support': True,
          'volume_backend_name': 'ceph',
          'replication_enabled': False,
          'allocated_capacity_gb': 1,
          'filter_function': None,
          'goodness_function': None,
          'timestamp': '2025-06-09T13:29:43.286226'})
]


class TestVolumeSizePollster(base.BaseTestCase):
    def setUp(self):
        super().setUp()
        conf = service.prepare_service([], [])
        self.manager = manager.AgentManager(0, conf)
        self.pollster = cinder.VolumeSizePollster(conf)

    def test_volume_size_pollster(self):
        volume_size_samples = list(
            self.pollster.get_samples(self.manager, {}, resources=VOLUME_LIST))
        self.assertEqual(1, len(volume_size_samples))
        self.assertEqual('volume.size', volume_size_samples[0].name)
        self.assertEqual(1, volume_size_samples[0].volume)
        self.assertEqual('6824974c08974d4db864bbaa6bc08303',
                         volume_size_samples[0].project_id)
        self.assertEqual('d94c18fb-b680-4912-9741-da69ee83c94f',
                         volume_size_samples[0].resource_id)
        self.assertEqual('f0019ee3-523c-45ab-b0b6-3adc529673e7',
                         volume_size_samples[0].resource_metadata["image_id"])
        self.assertEqual('1ae69721-d071-4156-a2bd-b11bb43ec2e3',
                         volume_size_samples[0].resource_metadata
                         ["instance_id"])
        self.assertEqual('nova', volume_size_samples[0].resource_metadata
                         ["availability_zone"])


class TestVolumeSnapshotSizePollster(base.BaseTestCase):
    def setUp(self):
        super().setUp()
        conf = service.prepare_service([], [])
        self.manager = manager.AgentManager(0, conf)
        self.pollster = cinder.VolumeSnapshotSize(conf)

    def test_volume_snapshot_size_pollster(self):
        volume_snapshot_size_samples = list(
            self.pollster.get_samples(
                self.manager, {}, resources=SNAPSHOT_LIST))
        self.assertEqual(1, len(volume_snapshot_size_samples))
        self.assertEqual('volume.snapshot.size',
                         volume_snapshot_size_samples[0].name)
        self.assertEqual(1, volume_snapshot_size_samples[0].volume)
        self.assertEqual('be255bd31eb944578000fc762fde6dcf',
                         volume_snapshot_size_samples[0].user_id)
        self.assertEqual('6824974c08974d4db864bbaa6bc08303',
                         volume_snapshot_size_samples[0].project_id)
        self.assertEqual('b1ea6783-f952-491e-a4ed-23a6a562e1cf',
                         volume_snapshot_size_samples[0].resource_id)


class TestVolumeBackupSizePollster(base.BaseTestCase):
    def setUp(self):
        super().setUp()
        conf = service.prepare_service([], [])
        self.manager = manager.AgentManager(0, conf)
        self.pollster = cinder.VolumeBackupSize(conf)

    def test_volume_backup_size_pollster(self):
        volume_backup_size_samples = list(
            self.pollster.get_samples(self.manager, {}, resources=BACKUP_LIST))
        self.assertEqual(1, len(volume_backup_size_samples))
        self.assertEqual('volume.backup.size',
                         volume_backup_size_samples[0].name)
        self.assertEqual(1, volume_backup_size_samples[0].volume)
        self.assertEqual('75a52125-85ff-4a8d-b2aa-580f3b22273f',
                         volume_backup_size_samples[0].resource_id)


class TestVolumeProviderPoolCapacityTotalPollster(base.BaseTestCase):
    def setUp(self):
        super().setUp()
        conf = service.prepare_service([], [])
        self.manager = manager.AgentManager(0, conf)
        self.pollster = cinder.VolumeProviderPoolCapacityTotal(conf)

    def test_volume_provider_pool_capacity_total_pollster(self):
        volume_pool_size_total_samples = list(
            self.pollster.get_samples(self.manager, {}, resources=POOL_LIST))
        self.assertEqual(2, len(volume_pool_size_total_samples))

        self.assertEqual('volume.provider.pool.capacity.total',
                         volume_pool_size_total_samples[0].name)
        self.assertEqual(28.5, volume_pool_size_total_samples[0].volume)
        self.assertEqual('localhost.localdomain@lvmdriver-1#lvmdriver-1',
                         volume_pool_size_total_samples[0].resource_id)

        self.assertEqual('volume.provider.pool.capacity.total',
                         volume_pool_size_total_samples[1].name)
        self.assertEqual(85.0, volume_pool_size_total_samples[1].volume)
        self.assertEqual('cinder-3ceee-volume-ceph-0@ceph#ceph',
                         volume_pool_size_total_samples[1].resource_id)


class TestVolumeProviderPoolCapacityFreePollster(base.BaseTestCase):
    def setUp(self):
        super().setUp()
        conf = service.prepare_service([], [])
        self.manager = manager.AgentManager(0, conf)
        self.pollster = cinder.VolumeProviderPoolCapacityFree(conf)

    def test_volume_provider_pool_capacity_free_pollster(self):
        volume_pool_size_free_samples = list(
            self.pollster.get_samples(self.manager, {}, resources=POOL_LIST))
        self.assertEqual(2, len(volume_pool_size_free_samples))

        self.assertEqual('volume.provider.pool.capacity.free',
                         volume_pool_size_free_samples[0].name)
        self.assertEqual(28.39, volume_pool_size_free_samples[0].volume)
        self.assertEqual('localhost.localdomain@lvmdriver-1#lvmdriver-1',
                         volume_pool_size_free_samples[0].resource_id)

        self.assertEqual('volume.provider.pool.capacity.free',
                         volume_pool_size_free_samples[1].name)
        self.assertEqual(85.0, volume_pool_size_free_samples[1].volume)
        self.assertEqual('cinder-3ceee-volume-ceph-0@ceph#ceph',
                         volume_pool_size_free_samples[1].resource_id)


class TestVolumeProviderPoolCapacityProvisionedPollster(base.BaseTestCase):
    def setUp(self):
        super().setUp()
        conf = service.prepare_service([], [])
        self.manager = manager.AgentManager(0, conf)
        self.pollster = cinder.VolumeProviderPoolCapacityProvisioned(conf)

    def test_volume_provider_pool_capacity_provisioned_pollster(self):
        volume_pool_size_provisioned_samples = list(
            self.pollster.get_samples(self.manager, {}, resources=POOL_LIST))
        self.assertEqual(1, len(volume_pool_size_provisioned_samples))
        self.assertEqual('volume.provider.pool.capacity.provisioned',
                         volume_pool_size_provisioned_samples[0].name)
        self.assertEqual(4.0, volume_pool_size_provisioned_samples[0].volume)
        self.assertEqual('localhost.localdomain@lvmdriver-1#lvmdriver-1',
                         volume_pool_size_provisioned_samples[0].resource_id)


class TestVolumeProviderPoolCapacityVirtualFreePollster(base.BaseTestCase):
    def setUp(self):
        super().setUp()
        conf = service.prepare_service([], [])
        self.manager = manager.AgentManager(0, conf)
        self.pollster = cinder.VolumeProviderPoolCapacityVirtualFree(conf)

    def test_volume_provider_pool_capacity_virtual_free_pollster(self):
        volume_pool_size_virtual_free_samples = list(
            self.pollster.get_samples(self.manager, {}, resources=POOL_LIST))
        self.assertEqual(1, len(volume_pool_size_virtual_free_samples))
        self.assertEqual('volume.provider.pool.capacity.virtual_free',
                         volume_pool_size_virtual_free_samples[0].name)
        self.assertEqual(566.0,
                         volume_pool_size_virtual_free_samples[0].volume)
        self.assertEqual('localhost.localdomain@lvmdriver-1#lvmdriver-1',
                         volume_pool_size_virtual_free_samples[0].resource_id)


class TestVolumeProviderPoolCapacityAllocatedPollster(base.BaseTestCase):
    def setUp(self):
        super().setUp()
        conf = service.prepare_service([], [])
        self.manager = manager.AgentManager(0, conf)
        self.pollster = cinder.VolumeProviderPoolCapacityAllocated(conf)

    def test_volume_provider_pool_capacity_allocated_pollster(self):
        volume_pool_size_allocated_samples = list(
            self.pollster.get_samples(self.manager, {}, resources=POOL_LIST))
        self.assertEqual(2, len(volume_pool_size_allocated_samples))

        self.assertEqual('volume.provider.pool.capacity.allocated',
                         volume_pool_size_allocated_samples[0].name)
        self.assertEqual(4, volume_pool_size_allocated_samples[0].volume)
        self.assertEqual('localhost.localdomain@lvmdriver-1#lvmdriver-1',
                         volume_pool_size_allocated_samples[0].resource_id)

        self.assertEqual('volume.provider.pool.capacity.allocated',
                         volume_pool_size_allocated_samples[1].name)
        self.assertEqual(1, volume_pool_size_allocated_samples[1].volume)
        self.assertEqual('cinder-3ceee-volume-ceph-0@ceph#ceph',
                         volume_pool_size_allocated_samples[1].resource_id)



#
# Copyright 2013 IBM Corp
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import json
import logging
import logging.handlers

from oslo_log import log
from urllib import parse as urlparse

from ceilometer import publisher

LOG = log.getLogger(__name__)


class FilePublisher(publisher.ConfigPublisherBase):
    """Publisher metering data to file.

    The file publisher pushes metering data into a file. The file name and
    location should be configured in ceilometer pipeline configuration file.
    If a file name and location is not specified, this File Publisher will not
    log any meters other than log a warning in Ceilometer log file.

    To enable this publisher, add the following section to the
    /etc/ceilometer/pipeline.yaml file or simply add it to an existing
    pipeline::

        -
            name: meter_file
            meters:
                - "*"
            publishers:
                - file:///var/test?max_bytes=10000000&backup_count=5&json

    File path is required for this publisher to work properly. If max_bytes
    or backup_count is missing, FileHandler will be used to save the metering
    data. If max_bytes and backup_count are present, RotatingFileHandler will
    be used to save the metering data. The json argument is used to explicitly
    ask ceilometer to write json into the file.
    """

    def __init__(self, conf, parsed_url):
        super().__init__(conf, parsed_url)

        self.publisher_logger = None
        path = parsed_url.path
        if not path:
            LOG.error('The path for the file publisher is required')
            return

        rfh = None
        max_bytes = 0
        backup_count = 0
        self.output_json = None
        # Handling other configuration options in the query string
        if parsed_url.query:
            params = urlparse.parse_qs(parsed_url.query,
                                       keep_blank_values=True)
            if "json" in params:
                self.output_json = True
            if params.get('max_bytes') and params.get('backup_count'):
                try:
                    max_bytes = int(params.get('max_bytes')[0])
                    backup_count = int(params.get('backup_count')[0])
                except ValueError:
                    LOG.error('max_bytes and backup_count should be '
                              'numbers.')
                    return
        # create rotating file handler
        rfh = logging.handlers.RotatingFileHandler(
            path, encoding='utf8', maxBytes=max_bytes,
            backupCount=backup_count)

        self.publisher_logger = logging.Logger('publisher.file')
        self.publisher_logger.propagate = False
        self.publisher_logger.setLevel(logging.INFO)
        rfh.setLevel(logging.INFO)
        self.publisher_logger.addHandler(rfh)

    def publish_samples(self, samples):
        """Send a metering message for publishing

        :param samples: Samples from pipeline after transformation
        """
        if self.publisher_logger:
            for sample in samples:
                if self.output_json:
                    self.publisher_logger.info(json.dumps(sample.as_dict()))
                else:
                    self.publisher_logger.info(sample.as_dict())

    def publish_events(self, events):
        """Send an event message for publishing

        :param events: events from pipeline after transformation
        """
        if self.publisher_logger:
            for event in events:
                if self.output_json:
                    self.publisher_logger.info(json.dumps(event.as_dict(),
                                                          default=str))
                else:
                    self.publisher_logger.info(event.as_dict())



#
# Copyright 2024 cmss, inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import json
import time

from oslo_log import log
from oslo_utils import timeutils

from ceilometer.publisher import http
from ceilometer import sample as smp

LOG = log.getLogger(__name__)


class OpentelemetryHttpPublisher(http.HttpPublisher):
    """Publish metering data to Opentelemetry Collector endpoint

    This dispatcher inherits from all options of the http dispatcher.

    To use this publisher for samples, add the following section to the
    /etc/ceilometer/pipeline.yaml file or simply add it to an existing
    pipeline::

          - name: meter_file
            meters:
                - "*"
            publishers:
                - opentelemetryhttp://opentelemetry-http-ip:4318/v1/metrics

    """

    HEADERS = {'Content-type': 'application/json'}

    @staticmethod
    def get_attribute_model(key, value):
        return {
            "key": key,
            "value": {
                "string_value": value
            }
        }

    def get_attributes_model(self, sample):
        attributes = []
        resource_id_attr = self.get_attribute_model("resource_id",
                                                    sample.resource_id)
        user_id_attr = self.get_attribute_model("user_id", sample.user_id)
        project_id_attr = self.get_attribute_model("project_id",
                                                   sample.project_id)

        attributes.append(resource_id_attr)
        attributes.append(user_id_attr)
        attributes.append(project_id_attr)

        return attributes

    @staticmethod
    def get_metrics_model(sample, data_points):
        name = sample.name.replace(".", "_")
        desc = str(sample.name) + " unit:" + sample.unit
        unit = sample.unit
        metrics = dict()
        metric_type = None
        if sample.type == smp.TYPE_CUMULATIVE:
            metric_type = "counter"
        else:
            metric_type = "gauge"
        metrics.update({
            "name": name,
            "description": desc,
            "unit": unit,
            metric_type: {"data_points": data_points}
        })
        return metrics

    @staticmethod
    def get_data_points_model(timestamp, attributes, volume):
        data_points = dict()
        struct_time = timeutils.parse_isotime(timestamp).timetuple()
        unix_time = int(time.mktime(struct_time))
        data_points.update({
            'attributes': attributes,
            "start_time_unix_nano": unix_time,
            "time_unix_nano": unix_time,
            "as_double": volume,
            "flags": 0
        })
        return data_points

    def get_data_model(self, sample, data_points):
        metrics = [self.get_metrics_model(sample, data_points)]
        data = {
            "resource_metrics": [{
                "scope_metrics": [{
                    "scope": {
                        "name": "ceilometer",
                        "version": "v1"
                    },
                    "metrics": metrics
                }]
            }]
        }
        return data

    def get_data_points(self, sample):
        # attributes contain basic metadata
        attributes = self.get_attributes_model(sample)
        try:
            return [self.get_data_points_model(
                sample.timestamp, attributes, sample.volume)]
        except Exception as e:
            LOG.warning("Get data point error, %s", e)
            return []

    def get_opentelemetry_model(self, sample):
        data_points = self.get_data_points(sample)
        if data_points:
            data = self.get_data_model(sample, data_points)
            return data
        else:
            return None

    def publish_samples(self, samples):
        """Send a metering message for publishing

        :param samples: Samples from pipeline after transformation
        """
        if not samples:
            LOG.warning('Data samples is empty!')
            return

        for s in samples:
            data = self.get_opentelemetry_model(s)
            if data:
                self._do_post(json.dumps(data))

    @staticmethod
    def publish_events(events):
        raise NotImplementedError



#
# Copyright 2012 Red Hat, Inc
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
"""Implementation of Inspector abstraction for libvirt."""

from lxml import etree
from oslo_log import log as logging
from oslo_utils import units

try:
    import libvirt
except ImportError:
    libvirt = None

from ceilometer.compute.pollsters import util
from ceilometer.compute.virt import inspector as virt_inspector
from ceilometer.compute.virt.libvirt import utils as libvirt_utils
from ceilometer.i18n import _

LOG = logging.getLogger(__name__)


class LibvirtInspector(virt_inspector.Inspector):

    def __init__(self, conf):
        super().__init__(conf)
        # NOTE(sileht): create a connection on startup
        self.connection
        self.cache = {}

    @property
    def connection(self):
        return libvirt_utils.refresh_libvirt_connection(self.conf, self)

    def _lookup_by_uuid(self, instance):
        instance_name = util.instance_name(instance)
        try:
            return self.connection.lookupByUUIDString(instance.id)
        except libvirt.libvirtError as ex:
            if libvirt_utils.is_disconnection_exception(ex):
                raise
            msg = _("Error from libvirt while looking up instance "
                    "<name=%(name)s, id=%(id)s>: "
                    "[Error Code %(error_code)s] "
                    "%(ex)s") % {'name': instance_name,
                                 'id': instance.id,
                                 'error_code': ex.get_error_code(),
                                 'ex': ex}
            raise virt_inspector.InstanceNotFoundException(msg)
        except Exception as ex:
            raise virt_inspector.InspectorException(str(ex))

    def _get_domain_not_shut_off_or_raise(self, instance):
        instance_name = util.instance_name(instance)
        domain = self._lookup_by_uuid(instance)

        state = domain.info()[0]
        if state == libvirt.VIR_DOMAIN_SHUTOFF:
            msg = _('Failed to inspect data of instance '
                    '<name=%(name)s, id=%(id)s>, '
                    'domain state is SHUTOFF.') % {
                'name': instance_name, 'id': instance.id}
            raise virt_inspector.InstanceShutOffException(msg)

        return domain

    @libvirt_utils.retry_on_disconnect
    def inspect_vnics(self, instance, duration):
        domain = self._get_domain_not_shut_off_or_raise(instance)

        tree = etree.fromstring(domain.XMLDesc(0))
        for iface in tree.findall('devices/interface'):
            target = iface.find('target')
            if target is not None:
                name = target.get('dev')
            else:
                continue
            mac = iface.find('mac')
            if mac is not None:
                mac_address = mac.get('address')
            else:
                continue
            fref = iface.find('filterref')
            if fref is not None:
                fref = fref.get('filter')

            params = {p.get('name').lower(): p.get('value')
                      for p in iface.findall('filterref/parameter')}

            # Extract interface ID
            try:
                interfaceid = iface.find('virtualport').find(
                    'parameters').get('interfaceid')
            except AttributeError:
                interfaceid = None

            # Extract source bridge
            try:
                bridge = iface.find('source').get('bridge')
            except AttributeError:
                bridge = None

            params['interfaceid'] = interfaceid
            params['bridge'] = bridge

            try:
                dom_stats = domain.interfaceStats(name)
            except libvirt.libvirtError as ex:
                LOG.warning("Error from libvirt when running instanceStats, "
                            "This may not be harmful, but please check : "
                            "%(ex)s",
                            {'ex': ex})
                continue

            # Retrieve previous values
            prev = self.cache.get(name)

            # Store values for next call
            self.cache[name] = dom_stats

            if prev:
                # Compute stats
                rx_delta = dom_stats[0] - prev[0]
                tx_delta = dom_stats[4] - prev[4]

                # Avoid negative values
                if rx_delta < 0:
                    rx_delta = dom_stats[0]
                if tx_delta < 0:
                    tx_delta = dom_stats[4]
            else:
                LOG.debug('No delta meter predecessor for %s / %s',
                          instance.id, name)
                rx_delta = 0
                tx_delta = 0

            yield virt_inspector.InterfaceStats(name=name,
                                                mac=mac_address,
                                                fref=fref,
                                                parameters=params,
                                                rx_bytes=dom_stats[0],
                                                rx_packets=dom_stats[1],
                                                rx_errors=dom_stats[2],
                                                rx_drop=dom_stats[3],
                                                rx_bytes_delta=rx_delta,
                                                tx_bytes=dom_stats[4],
                                                tx_packets=dom_stats[5],
                                                tx_errors=dom_stats[6],
                                                tx_drop=dom_stats[7],
                                                tx_bytes_delta=tx_delta)

    @staticmethod
    def _get_disk_devices(domain):
        tree = etree.fromstring(domain.XMLDesc(0))
        return filter(bool, [target.get("dev") for target in
                             tree.findall('devices/disk/target')
                             if target.getparent().find('source') is not None])

    @libvirt_utils.retry_on_disconnect
    def inspect_disks(self, instance, duration):
        domain = self._get_domain_not_shut_off_or_raise(instance)
        for device in self._get_disk_devices(domain):
            try:
                block_stats = domain.blockStats(device)
                block_stats_flags = domain.blockStatsFlags(device, 0)
                yield virt_inspector.DiskStats(
                    device=device,
                    read_requests=block_stats[0], read_bytes=block_stats[1],
                    write_requests=block_stats[2], write_bytes=block_stats[3],
                    errors=block_stats[4],
                    wr_total_times=block_stats_flags['wr_total_times'],
                    rd_total_times=block_stats_flags['rd_total_times'])
            except libvirt.libvirtError as ex:
                # raised error even if lock is acquired while live migration,
                # even it looks normal.
                LOG.warning("Error from libvirt while checking blockStats, "
                            "This may not be harmful, but please check : "
                            "%(ex)s", {'ex': ex})
                pass

    @libvirt_utils.retry_on_disconnect
    def inspect_disk_info(self, instance, duration):
        domain = self._get_domain_not_shut_off_or_raise(instance)
        for device in self._get_disk_devices(domain):
            block_info = domain.blockInfo(device)
            # if vm mount cdrom, libvirt will align by 4K bytes, capacity may
            # be smaller than physical, avoid with this.
            # https://libvirt.org/html/libvirt-libvirt-domain.html
            disk_capacity = max(block_info[0], block_info[2])
            yield virt_inspector.DiskInfo(device=device,
                                          capacity=disk_capacity,
                                          allocation=block_info[1],
                                          physical=block_info[2])

    @libvirt_utils.raise_nodata_if_unsupported
    @libvirt_utils.retry_on_disconnect
    def inspect_instance(self, instance, duration=None):
        domain = self._get_domain_not_shut_off_or_raise(instance)

        memory_actual = None
        memory_available = None
        memory_used = memory_resident = None
        memory_swap_in = memory_swap_out = None
        memory_stats = domain.memoryStats()
        # Stat provided from libvirt is in KiB, converting it to MiB.
        if 'actual' in memory_stats:
            memory_actual = memory_stats['actual'] / units.Ki
        if 'available' in memory_stats:
            memory_available = memory_stats['available'] / units.Ki
        if 'usable' in memory_stats and 'available' in memory_stats:
            memory_used = (memory_stats['available'] -
                           memory_stats['usable']) / units.Ki
        elif 'available' in memory_stats and 'unused' in memory_stats:
            memory_used = (memory_stats['available'] -
                           memory_stats['unused']) / units.Ki
        if 'rss' in memory_stats:
            memory_resident = memory_stats['rss'] / units.Ki
        if 'swap_in' in memory_stats and 'swap_out' in memory_stats:
            memory_swap_in = memory_stats['swap_in'] / units.Ki
            memory_swap_out = memory_stats['swap_out'] / units.Ki

        # TODO(sileht): stats also have the disk/vnic info
        # we could use that instead of the old method for Queen
        stats = self.connection.domainListGetStats([domain], 0)[0][1]
        cpu_time = 0
        current_cpus = stats.get('vcpu.current')
        # Iterate over the maximum number of CPUs here, and count the
        # actual number encountered, since the vcpu.x structure can
        # have holes according to
        # https://libvirt.org/git/?p=libvirt.git;a=blob;f=src/libvirt-domain.c
        # virConnectGetAllDomainStats()
        for vcpu in range(stats.get('vcpu.maximum', 0)):
            try:
                cpu_time += (stats.get('vcpu.%s.time' % vcpu) +
                             stats.get('vcpu.%s.wait' % vcpu))
                current_cpus -= 1
            except TypeError:
                # pass here, if there are too many holes, the cpu count will
                # not match, so don't need special error handling.
                pass

        if current_cpus:
            # There wasn't enough data, so fall back
            cpu_time = stats.get('cpu.time')

        return virt_inspector.InstanceStats(
            power_state=domain.info()[0],
            cpu_number=stats.get('vcpu.current'),
            cpu_time=cpu_time,
            memory_actual=memory_actual,
            memory_available=memory_available,
            memory_usage=memory_used,
            memory_resident=memory_resident,
            memory_swap_in=memory_swap_in,
            memory_swap_out=memory_swap_out,
            cpu_cycles=stats.get("perf.cpu_cycles"),
            instructions=stats.get("perf.instructions"),
            cache_references=stats.get("perf.cache_references"),
            cache_misses=stats.get("perf.cache_misses")
        )



#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import traceback
import uuid

import fixtures
from oslo_utils import timeutils

from ceilometer.event import models
from ceilometer.pipeline import base as pipeline
from ceilometer.pipeline import event
from ceilometer import publisher
from ceilometer.publisher import test as test_publisher
from ceilometer import service
from ceilometer.tests import base


class EventPipelineTestCase(base.BaseTestCase):

    def get_publisher(self, conf, url, namespace=''):
        fake_drivers = {'test://': test_publisher.TestPublisher,
                        'new://': test_publisher.TestPublisher,
                        'except://': self.PublisherClassException}
        return fake_drivers[url](conf, url)

    class PublisherClassException(publisher.ConfigPublisherBase):
        def publish_samples(self, samples):
            pass

        def publish_events(self, events):
            raise Exception()

    def setUp(self):
        super().setUp()
        self.CONF = service.prepare_service([], [])

        self.test_event = models.Event(
            message_id=uuid.uuid4(),
            event_type='a',
            generated=timeutils.utcnow(),
            traits=[
                models.Trait('t_text', 1, 'text_trait'),
                models.Trait('t_int', 2, 'int_trait'),
                models.Trait('t_float', 3, 'float_trait'),
                models.Trait('t_datetime', 4, 'datetime_trait')
            ],
            raw={'status': 'started'}
        )

        self.test_event2 = models.Event(
            message_id=uuid.uuid4(),
            event_type='b',
            generated=timeutils.utcnow(),
            traits=[
                models.Trait('t_text', 1, 'text_trait'),
                models.Trait('t_int', 2, 'int_trait'),
                models.Trait('t_float', 3, 'float_trait'),
                models.Trait('t_datetime', 4, 'datetime_trait')
            ],
            raw={'status': 'stopped'}
        )

        self.useFixture(fixtures.MockPatchObject(
            publisher, 'get_publisher', side_effect=self.get_publisher))

        self._setup_pipeline_cfg()

        self._reraise_exception = True
        self.useFixture(fixtures.MockPatch(
            'ceilometer.pipeline.base.LOG.exception',
            side_effect=self._handle_reraise_exception))

    def _handle_reraise_exception(self, *args, **kwargs):
        if self._reraise_exception:
            raise Exception(traceback.format_exc())

    def _setup_pipeline_cfg(self):
        """Setup the appropriate form of pipeline config."""
        source = {'name': 'test_source',
                  'events': ['a'],
                  'sinks': ['test_sink']}
        sink = {'name': 'test_sink',
                'publishers': ['test://']}
        self.pipeline_cfg = {'sources': [source], 'sinks': [sink]}

    def _augment_pipeline_cfg(self):
        """Augment the pipeline config with an additional element."""
        self.pipeline_cfg['sources'].append({
            'name': 'second_source',
            'events': ['b'],
            'sinks': ['second_sink']
        })
        self.pipeline_cfg['sinks'].append({
            'name': 'second_sink',
            'publishers': ['new://'],
        })

    def _break_pipeline_cfg(self):
        """Break the pipeline config with a malformed element."""
        self.pipeline_cfg['sources'].append({
            'name': 'second_source',
            'events': ['b'],
            'sinks': ['second_sink']
        })
        self.pipeline_cfg['sinks'].append({
            'name': 'second_sink',
            'publishers': ['except'],
        })

    def _dup_pipeline_name_cfg(self):
        """Break the pipeline config with duplicate pipeline name."""
        self.pipeline_cfg['sources'].append({
            'name': 'test_source',
            'events': ['a'],
            'sinks': ['test_sink']
        })

    def _set_pipeline_cfg(self, field, value):
        if field in self.pipeline_cfg['sources'][0]:
            self.pipeline_cfg['sources'][0][field] = value
        else:
            self.pipeline_cfg['sinks'][0][field] = value

    def _extend_pipeline_cfg(self, field, value):
        if field in self.pipeline_cfg['sources'][0]:
            self.pipeline_cfg['sources'][0][field].extend(value)
        else:
            self.pipeline_cfg['sinks'][0][field].extend(value)

    def _unset_pipeline_cfg(self, field):
        if field in self.pipeline_cfg['sources'][0]:
            del self.pipeline_cfg['sources'][0][field]
        else:
            del self.pipeline_cfg['sinks'][0][field]

    def _build_and_set_new_pipeline(self):
        name = self.cfg2file(self.pipeline_cfg)
        self.CONF.set_override('event_pipeline_cfg_file', name)

    def _exception_create_pipelinemanager(self):
        self._build_and_set_new_pipeline()
        self.assertRaises(pipeline.PipelineException,
                          event.EventPipelineManager, self.CONF)

    def test_no_events(self):
        self._unset_pipeline_cfg('events')
        self._exception_create_pipelinemanager()

    def test_no_name(self):
        self._unset_pipeline_cfg('name')
        self._exception_create_pipelinemanager()

    def test_name(self):
        self._build_and_set_new_pipeline()
        pipeline_manager = event.EventPipelineManager(self.CONF)
        for pipe in pipeline_manager.pipelines:
            self.assertTrue(pipe.name.startswith('event:'))

    def test_no_publishers(self):
        self._unset_pipeline_cfg('publishers')
        self._exception_create_pipelinemanager()

    def test_check_events_include_exclude_same(self):
        event_cfg = ['a', '!a']
        self._set_pipeline_cfg('events', event_cfg)
        self._exception_create_pipelinemanager()

    def test_check_events_include_exclude(self):
        event_cfg = ['a', '!b']
        self._set_pipeline_cfg('events', event_cfg)
        self._exception_create_pipelinemanager()

    def test_check_events_wildcard_included(self):
        event_cfg = ['a', '*']
        self._set_pipeline_cfg('events', event_cfg)
        self._exception_create_pipelinemanager()

    def test_check_publishers_invalid_publisher(self):
        publisher_cfg = ['test_invalid']
        self._set_pipeline_cfg('publishers', publisher_cfg)

    def test_multiple_included_events(self):
        event_cfg = ['a', 'b']
        self._set_pipeline_cfg('events', event_cfg)
        self._build_and_set_new_pipeline()
        pipeline_manager = event.EventPipelineManager(self.CONF)

        with pipeline_manager.publisher() as p:
            p([self.test_event])

        publisher = pipeline_manager.pipelines[0].publishers[0]
        self.assertEqual(1, len(publisher.events))

        with pipeline_manager.publisher() as p:
            p([self.test_event2])

        self.assertEqual(2, len(publisher.events))
        self.assertEqual('a', getattr(publisher.events[0], 'event_type'))
        self.assertEqual('b', getattr(publisher.events[1], 'event_type'))

    def test_event_non_match(self):
        event_cfg = ['nomatch']
        self._set_pipeline_cfg('events', event_cfg)
        self._build_and_set_new_pipeline()
        pipeline_manager = event.EventPipelineManager(self.CONF)
        with pipeline_manager.publisher() as p:
            p([self.test_event])

        publisher = pipeline_manager.pipelines[0].publishers[0]
        self.assertEqual(0, len(publisher.events))
        self.assertEqual(0, publisher.calls)

    def test_wildcard_event(self):
        event_cfg = ['*']
        self._set_pipeline_cfg('events', event_cfg)
        self._build_and_set_new_pipeline()
        pipeline_manager = event.EventPipelineManager(self.CONF)
        with pipeline_manager.publisher() as p:
            p([self.test_event])

        publisher = pipeline_manager.pipelines[0].publishers[0]
        self.assertEqual(1, len(publisher.events))
        self.assertEqual('a', getattr(publisher.events[0], 'event_type'))

    def test_wildcard_excluded_events(self):
        event_cfg = ['*', '!a']
        self._set_pipeline_cfg('events', event_cfg)
        self._build_and_set_new_pipeline()
        pipeline_manager = event.EventPipelineManager(self.CONF)
        pipe = pipeline_manager.pipelines[0]
        self.assertFalse(pipe.source.support_event('a'))

    def test_wildcard_excluded_events_not_excluded(self):
        event_cfg = ['*', '!b']
        self._set_pipeline_cfg('events', event_cfg)
        self._build_and_set_new_pipeline()
        pipeline_manager = event.EventPipelineManager(self.CONF)
        with pipeline_manager.publisher() as p:
            p([self.test_event])
        publisher = pipeline_manager.pipelines[0].publishers[0]
        self.assertEqual(1, len(publisher.events))
        self.assertEqual('a', getattr(publisher.events[0], 'event_type'))

    def test_all_excluded_events_not_excluded(self):
        event_cfg = ['!b', '!c']
        self._set_pipeline_cfg('events', event_cfg)
        self._build_and_set_new_pipeline()
        pipeline_manager = event.EventPipelineManager(self.CONF)
        with pipeline_manager.publisher() as p:
            p([self.test_event])

        publisher = pipeline_manager.pipelines[0].publishers[0]
        self.assertEqual(1, len(publisher.events))
        self.assertEqual('a', getattr(publisher.events[0], 'event_type'))

    def test_all_excluded_events_excluded(self):
        event_cfg = ['!a', '!c']
        self._set_pipeline_cfg('events', event_cfg)
        self._build_and_set_new_pipeline()
        pipeline_manager = event.EventPipelineManager(self.CONF)
        pipe = pipeline_manager.pipelines[0]
        self.assertFalse(pipe.source.support_event('a'))
        self.assertTrue(pipe.source.support_event('b'))
        self.assertFalse(pipe.source.support_event('c'))

    def test_wildcard_and_excluded_wildcard_events(self):
        event_cfg = ['*', '!compute.*']
        self._set_pipeline_cfg('events', event_cfg)
        self._build_and_set_new_pipeline()
        pipeline_manager = event.EventPipelineManager(self.CONF)
        pipe = pipeline_manager.pipelines[0]
        self.assertFalse(pipe.source.
                         support_event('compute.instance.create.start'))
        self.assertTrue(pipe.source.support_event('identity.user.create'))

    def test_included_event_and_wildcard_events(self):
        event_cfg = ['compute.instance.create.start', 'identity.*']
        self._set_pipeline_cfg('events', event_cfg)
        self._build_and_set_new_pipeline()
        pipeline_manager = event.EventPipelineManager(self.CONF)
        pipe = pipeline_manager.pipelines[0]
        self.assertTrue(pipe.source.support_event('identity.user.create'))
        self.assertTrue(pipe.source.
                        support_event('compute.instance.create.start'))
        self.assertFalse(pipe.source.
                         support_event('compute.instance.create.stop'))

    def test_excluded_event_and_excluded_wildcard_events(self):
        event_cfg = ['!compute.instance.create.start', '!identity.*']
        self._set_pipeline_cfg('events', event_cfg)
        self._build_and_set_new_pipeline()
        pipeline_manager = event.EventPipelineManager(self.CONF)
        pipe = pipeline_manager.pipelines[0]
        self.assertFalse(pipe.source.support_event('identity.user.create'))
        self.assertFalse(pipe.source.
                         support_event('compute.instance.create.start'))
        self.assertTrue(pipe.source.
                        support_event('compute.instance.create.stop'))

    def test_multiple_pipeline(self):
        self._augment_pipeline_cfg()
        self._build_and_set_new_pipeline()
        pipeline_manager = event.EventPipelineManager(self.CONF)
        with pipeline_manager.publisher() as p:
            p([self.test_event, self.test_event2])

        publisher = pipeline_manager.pipelines[0].publishers[0]
        self.assertEqual(1, len(publisher.events))
        self.assertEqual(1, publisher.calls)
        self.assertEqual('a', getattr(publisher.events[0], 'event_type'))
        new_publisher = pipeline_manager.pipelines[1].publishers[0]
        self.assertEqual(1, len(new_publisher.events))
        self.assertEqual(1, new_publisher.calls)
        self.assertEqual('b', getattr(new_publisher.events[0], 'event_type'))

    def test_multiple_publisher(self):
        self._set_pipeline_cfg('publishers', ['test://', 'new://'])
        self._build_and_set_new_pipeline()
        pipeline_manager = event.EventPipelineManager(self.CONF)

        with pipeline_manager.publisher() as p:
            p([self.test_event])

        publisher = pipeline_manager.pipelines[0].publishers[0]
        new_publisher = pipeline_manager.pipelines[0].publishers[1]
        self.assertEqual(1, len(publisher.events))
        self.assertEqual(1, len(new_publisher.events))
        self.assertEqual('a', getattr(new_publisher.events[0], 'event_type'))
        self.assertEqual('a', getattr(publisher.events[0], 'event_type'))

    def test_multiple_publisher_isolation(self):
        self._reraise_exception = False
        self._set_pipeline_cfg('publishers', ['except://', 'new://'])
        self._build_and_set_new_pipeline()
        pipeline_manager = event.EventPipelineManager(self.CONF)
        with pipeline_manager.publisher() as p:
            p([self.test_event])

        publisher = pipeline_manager.pipelines[0].publishers[1]
        self.assertEqual(1, len(publisher.events))
        self.assertEqual('a', getattr(publisher.events[0], 'event_type'))

    def test_unique_pipeline_names(self):
        self._dup_pipeline_name_cfg()
        self._exception_create_pipelinemanager()



# Copyright 2014 Mirantis, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import collections
from time import monotonic as now

from oslo_log import log
from oslo_utils import timeutils

import ceilometer
from ceilometer.compute.pollsters import util
from ceilometer.compute.virt import inspector as virt_inspector
from ceilometer.polling import plugin_base
from ceilometer import sample

LOG = log.getLogger(__name__)


class NoVolumeException(Exception):
    pass


class GenericComputePollster(plugin_base.PollsterBase):
    """This class aims to cache instance statistics data

    First polled pollsters that inherit of this will retrieve and cache
    stats of an instance, then other pollsters will just build the samples
    without queyring the backend anymore.
    """

    sample_name = None
    sample_unit = ''
    sample_type = sample.TYPE_GAUGE
    sample_stats_key = None
    inspector_method = None

    def setup_environment(self):
        super().setup_environment()
        self.inspector = GenericComputePollster._get_inspector(self.conf)

    @staticmethod
    def aggregate_method(stats):
        # Don't aggregate anything by default
        return stats

    @staticmethod
    def _get_inspector(conf):
        # FIXME(sileht): This doesn't looks threadsafe...
        try:
            inspector = GenericComputePollster._inspector
        except AttributeError:
            inspector = virt_inspector.get_hypervisor_inspector(conf)
            GenericComputePollster._inspector = inspector
        return inspector

    @property
    def default_discovery(self):
        return 'local_instances'

    def _record_poll_time(self):
        """Method records current time as the poll time.

        :return: time in seconds since the last poll time was recorded
        """
        current_time = timeutils.utcnow()
        duration = None
        if hasattr(self, '_last_poll_time'):
            duration = timeutils.delta_seconds(self._last_poll_time,
                                               current_time)
        self._last_poll_time = current_time
        return duration

    @staticmethod
    def get_additional_metadata(instance, stats):
        pass

    @staticmethod
    def get_resource_id(instance, stats):
        return instance.id

    def _inspect_cached(self, cache, instance, duration):
        cache.setdefault(self.inspector_method, {})
        if instance.id not in cache[self.inspector_method]:
            result = getattr(self.inspector, self.inspector_method)(
                instance, duration)
            polled_time = now()
            # Ensure we don't cache an iterator
            if isinstance(result, collections.abc.Iterable):
                result = list(result)
            else:
                result = [result]
            cache[self.inspector_method][instance.id] = (polled_time, result)
        return cache[self.inspector_method][instance.id]

    def _stats_to_sample(self, instance, stats, polled_time):
        volume = getattr(stats, self.sample_stats_key)
        LOG.debug(
            "%(instance_id)s/%(name)s volume: %(volume)s",
            {'name': self.sample_name,
             'instance_id': instance.id,
             'volume': (volume if volume is not None else 'Unavailable')})

        if volume is None:
            raise NoVolumeException()

        return util.make_sample_from_instance(
            self.conf,
            instance,
            name=self.sample_name,
            unit=self.sample_unit,
            type=self.sample_type,
            resource_id=self.get_resource_id(instance, stats),
            volume=volume,
            additional_metadata=self.get_additional_metadata(
                instance, stats),
            monotonic_time=polled_time,
        )

    def get_samples(self, manager, cache, resources):
        self._inspection_duration = self._record_poll_time()
        for instance in resources:
            try:
                polled_time, result = self._inspect_cached(
                    cache, instance, self._inspection_duration)
                if not result:
                    continue
                for stats in self.aggregate_method(result):
                    yield self._stats_to_sample(instance, stats, polled_time)
            except NoVolumeException:
                # FIXME(sileht): This should be a removed... but I will
                # not change the test logic for now
                LOG.warning("%(name)s statistic in not available for "
                            "instance %(instance_id)s",
                            {'name': self.sample_name,
                             'instance_id': instance.id})
            except virt_inspector.InstanceNotFoundException as err:
                # Instance was deleted while getting samples. Ignore it.
                LOG.debug('Exception while getting samples %s', err)
            except virt_inspector.InstanceShutOffException as e:
                LOG.debug('Instance %(instance_id)s was shut off while '
                          'getting sample of %(name)s: %(exc)s',
                          {'instance_id': instance.id,
                           'name': self.sample_name, 'exc': e})
            except virt_inspector.NoDataException as e:
                LOG.warning('Cannot inspect data of %(pollster)s for '
                            '%(instance_id)s, non-fatal reason: %(exc)s',
                            {'pollster': self.__class__.__name__,
                             'instance_id': instance.id, 'exc': e})
            except ceilometer.NotImplementedError:
                # Selected inspector does not implement this pollster.
                LOG.debug('%(inspector)s does not provide data for '
                          '%(pollster)s',
                          {'inspector': self.inspector.__class__.__name__,
                           'pollster': self.__class__.__name__})
                raise plugin_base.PollsterPermanentError(resources)
            except Exception as err:
                LOG.error(
                    'Could not get %(name)s events for %(id)s: %(e)s', {
                        'name': self.sample_name, 'id': instance.id, 'e': err},
                    exc_info=True)


class InstanceMetadataPollster(plugin_base.PollsterBase):
    """A base class for implementing a pollster using instance metadata.

    This metadata is originally supplied by Nova, but if
    instance_discovery_method is set to libvirt_metadata,
    metadata is fetched from the local libvirt socket,
    just like with the standard compute pollsters.
    """

    sample_name = None
    sample_unit = ''
    sample_type = sample.TYPE_GAUGE

    @property
    def default_discovery(self):
        return 'local_instances'

    def get_resource_id(self, instance):
        return instance.id

    def get_volume(self, instance):
        raise ceilometer.NotImplementedError

    def get_additional_metadata(self, instance):
        return {}

    def get_samples(self, manager, cache, resources):
        for instance in resources:
            yield util.make_sample_from_instance(
                self.conf,
                instance,
                name=self.sample_name,
                unit=self.sample_unit,
                type=self.sample_type,
                resource_id=self.get_resource_id(instance),
                volume=self.get_volume(instance),
                additional_metadata=self.get_additional_metadata(instance),
                monotonic_time=now(),
            )



#
# Copyright 2013 Intel Corp.
# Copyright 2013-2014 eNovance
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import abc

from oslo_log import log
from oslo_utils import netutils
from stevedore import driver


LOG = log.getLogger(__name__)


def get_publisher(conf, url, namespace):
    """Get publisher driver and load it.

    :param url: URL for the publisher
    :param namespace: Namespace to use to look for drivers.
    """
    parse_result = netutils.urlsplit(url)
    loaded_driver = driver.DriverManager(namespace, parse_result.scheme)
    return loaded_driver.driver(conf, parse_result)


class ConfigPublisherBase(metaclass=abc.ABCMeta):
    """Base class for plugins that publish data."""

    def __init__(self, conf, parsed_url):
        self.conf = conf

    @abc.abstractmethod
    def publish_samples(self, samples):
        """Publish samples into final conduit."""

    @abc.abstractmethod
    def publish_events(self, events):
        """Publish events into final conduit."""



#
# Copyright 2022 Red Hat, Inc
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
"""Tests for ceilometer/polling/prom_exporter.py"""

from oslotest import base

from unittest import mock
from unittest.mock import call

from ceilometer.polling import manager
from ceilometer.polling import prom_exporter
from ceilometer import service


COUNTER_SOURCE = 'testsource'


class TestPromExporter(base.BaseTestCase):
    test_disk_latency = [
        {
            'source': 'openstack',
            'counter_name': 'disk.device.read.latency',
            'counter_type': 'cumulative',
            'counter_unit': 'ns',
            'counter_volume': 132128682,
            'user_id': '6e7d71415cd5401cbe103829c9c5dec2',
            'user_name': None,
            'project_id': 'd965489b7f894cbda89cd2e25bfd85a0',
            'project_name': None,
            'resource_id': 'e536fff6-b20d-4aa5-ac2f-d15ac8b3af63-vda',
            'timestamp': '2024-06-20T09:32:36.521082',
            'resource_metadata': {
                'display_name': 'myserver',
                'name': 'instance-00000002',
                'instance_id': 'e536fff6-b20d-4aa5-ac2f-d15ac8b3af63',
                'instance_type': 'tiny',
                'host': 'e0d297f5df3b62ec73c8d42b',
                'instance_host': 'devstack',
                'flavor': {
                    'id': '4af9ac72-5787-4f86-8644-0faa87ce7c83',
                    'name': 'tiny',
                    'vcpus': 1,
                    'ram': 512,
                    'disk': 1,
                    'ephemeral': 0,
                    'swap': 0
                },
                'status': 'active',
                'state': 'running',
                'task_state': '',
                'image': {
                    'id': '71860ed5-f66d-43e0-9514-f1d188106284'
                },
                'image_ref': '71860ed5-f66d-43e0-9514-f1d188106284',
                'image_ref_url': None,
                'architecture': 'x86_64',
                'os_type': 'hvm',
                'vcpus': 1,
                'memory_mb': 512,
                'disk_gb': 1,
                'ephemeral_gb': 0,
                'root_gb': 1,
                'disk_name': 'vda',
                'user_metadata': {
                    'custom_label': 'custom value'
                }
            },
            'message_id': '078029c7-2ee8-11ef-a915-bd45e2085de3',
            'monotonic_time': 1819980.112406547,
            'message_signature': 'f8d9a411b0cd0cb0d34e83'
        },
        {
            'source': 'openstack',
            'counter_name': 'disk.device.read.latency',
            'counter_type': 'cumulative',
            'counter_unit': 'ns',
            'counter_volume': 232128754,
            'user_id': '6e7d71415cd5401cbe103829c9c5dec2',
            'user_name': None,
            'project_id': 'd965489b7f894cbda89cd2e25bfd85a0',
            'project_name': None,
            'resource_id': 'e536fff6-b20d-4aa5-ac2f-d15ac8b3af63-vda',
            'timestamp': '2024-06-20T09:32:46.521082',
            'resource_metadata': {
                'display_name': 'myserver',
                'name': 'instance-00000002',
                'instance_id': 'e536fff6-b20d-4aa5-ac2f-d15ac8b3af63',
                'instance_type': 'tiny',
                'host': 'e0d297f5df3b62ec73c8d42b',
                'instance_host': 'devstack',
                'flavor': {
                    'id': '4af9ac72-5787-4f86-8644-0faa87ce7c83',
                    'name': 'tiny',
                    'vcpus': 1,
                    'ram': 512,
                    'disk': 1,
                    'ephemeral': 0,
                    'swap': 0
                },
                'status': 'active',
                'state': 'running',
                'task_state': '',
                'image': {
                    'id': '71860ed5-f66d-43e0-9514-f1d188106284'
                },
                'image_ref': '71860ed5-f66d-43e0-9514-f1d188106284',
                'image_ref_url': None,
                'architecture': 'x86_64',
                'os_type': 'hvm',
                'vcpus': 1,
                'memory_mb': 512,
                'disk_gb': 1,
                'ephemeral_gb': 0,
                'root_gb': 1,
                'disk_name': 'vda',
                'user_metadata': {
                    'custom_label': 'custom value'
                }
            },
            'message_id': '078029c7-2ee8-11ef-a915-bd45e2085de4',
            'monotonic_time': 1819990.112406547,
            'message_signature': 'f8d9a411b0cd0cb0d34e84'
        }
    ]

    test_memory_usage = [
        {
            'source': 'openstack',
            'counter_name': 'memory.usage',
            'counter_type': 'gauge',
            'counter_unit': 'MiB',
            'counter_volume': 37.98046875,
            'user_id': '6e7d71415cd5401cbe103829c9c5dec2',
            'user_name': None,
            'project_id': 'd965489b7f894cbda89cd2e25bfd85a0',
            'project_name': None,
            'resource_id': 'e536fff6-b20d-4aa5-ac2f-d15ac8b3af63',
            'timestamp': '2024-06-20T09:32:36.515823',
            'resource_metadata': {
                'display_name': 'myserver',
                'name': 'instance-00000002',
                'instance_id': 'e536fff6-b20d-4aa5-ac2f-d15ac8b3af63',
                'instance_type': 'tiny',
                'host': 'e0d297f5df3b62ec73c8d42b',
                'instance_host': 'devstack',
                'flavor': {
                    'id': '4af9ac72-5787-4f86-8644-0faa87ce7c83',
                    'name': 'tiny',
                    'vcpus': 1,
                    'ram': 512,
                    'disk': 1,
                    'ephemeral': 0,
                    'swap': 0
                },
                'status': 'active',
                'state': 'running',
                'task_state': '',
                'image': {
                    'id': '71860ed5-f66d-43e0-9514-f1d188106284'
                },
                'image_ref': '71860ed5-f66d-43e0-9514-f1d188106284',
                'image_ref_url': None,
                'architecture': 'x86_64',
                'os_type': 'hvm',
                'vcpus': 1,
                'memory_mb': 512,
                'disk_gb': 1,
                'ephemeral_gb': 0,
                'root_gb': 1
            },
            'message_id': '078029bf-2ee8-11ef-a915-bd45e2085de3',
            'monotonic_time': 1819980.131767362,
            'message_signature': 'f8d9a411b0cd0cb0d34e83'
        }
    ]

    test_image_size = [
        {
            'source': 'openstack',
            'counter_name': 'image.size',
            'counter_type': 'gauge',
            'counter_unit': 'B',
            'counter_volume': 16344576,
            'user_id': None,
            'user_name': None,
            'project_id': 'd965489b7f894cbda89cd2e25bfd85a0',
            'project_name': None,
            'resource_id': 'f9276c96-8a12-432b-96a1-559d70715f97',
            'timestamp': '2024-06-20T09:40:17.118871',
            'resource_metadata': {
                'status': 'active',
                'visibility': 'public',
                'name': 'cirros2',
                'container_format': 'bare',
                'created_at': '2024-05-30T11:38:52Z',
                'disk_format': 'qcow2',
                'updated_at': '2024-05-30T11:38:52Z',
                'min_disk': 0,
                'protected': False,
                'checksum': '7734eb3945297adc90ddc6cebe8bb082',
                'min_ram': 0,
                'tags': [],
                'virtual_size': 117440512,
                'user_metadata': {
                    'server_group': 'server_group123'
                }
            },
            'message_id': '19f8f78a-2ee9-11ef-a95f-bd45e2085de3',
            'monotonic_time': None,
            'message_signature': 'f8d9a411b0cd0cb0d34e83'
        }
    ]

    @mock.patch('ceilometer.polling.prom_exporter.export')
    def test_prom_disabled(self, export):
        CONF = service.prepare_service([], [])
        manager.AgentManager(0, CONF)

        export.assert_not_called()

    @mock.patch('ceilometer.polling.prom_exporter.export')
    def test_export_called(self, export):
        CONF = service.prepare_service([], [])
        CONF.set_override('enable_prometheus_exporter', True, group='polling')
        CONF.set_override('prometheus_listen_addresses', [
            '127.0.0.1:9101',
            '127.0.0.1:9102',
            '[::1]:9103',
            'localhost:9104',
        ], group='polling')
        manager.AgentManager(0, CONF)

        export.assert_has_calls([
            call('127.0.0.1', 9101, None, None),
            call('127.0.0.1', 9102, None, None),
            call('::1', 9103, None, None),
            call('localhost', 9104, None, None),
        ])

    @mock.patch('ceilometer.polling.prom_exporter.export')
    def test_export_called_tls_disabled(self, export):
        CONF = service.prepare_service([], [])
        CONF.set_override('enable_prometheus_exporter', True, group='polling')
        CONF.set_override('prometheus_tls_enable', False, group='polling')
        CONF.set_override('prometheus_tls_certfile', "cert.pem",
                          group='polling')
        CONF.set_override('prometheus_listen_addresses', [
            '127.0.0.1:9101',
            '127.0.0.1:9102',
            '[::1]:9103',
            'localhost:9104',
        ], group='polling')
        manager.AgentManager(0, CONF)

        export.assert_has_calls([
            call('127.0.0.1', 9101, None, None),
            call('127.0.0.1', 9102, None, None),
            call('::1', 9103, None, None),
            call('localhost', 9104, None, None),
        ])

    @mock.patch('ceilometer.polling.prom_exporter.export')
    def test_export_called_with_tls(self, export):
        CONF = service.prepare_service([], [])
        CONF.set_override('enable_prometheus_exporter', True, group='polling')
        CONF.set_override('prometheus_listen_addresses', [
            '127.0.0.1:9101',
            '127.0.0.1:9102',
            '[::1]:9103',
            'localhost:9104',
        ], group='polling')
        CONF.set_override('prometheus_tls_enable', True, group='polling')
        CONF.set_override('prometheus_tls_certfile', "cert.pem",
                          group='polling')
        CONF.set_override('prometheus_tls_keyfile', "key.pem",
                          group='polling')
        manager.AgentManager(0, CONF)

        export.assert_has_calls([
            call('127.0.0.1', 9101, "cert.pem", "key.pem"),
            call('127.0.0.1', 9102, "cert.pem", "key.pem"),
            call('::1', 9103, "cert.pem", "key.pem"),
            call('localhost', 9104, "cert.pem", "key.pem"),
        ])

    @mock.patch('ceilometer.polling.prom_exporter.export')
    def test_export_fails_if_incomplete_tls(self, export):
        CONF = service.prepare_service([], [])
        CONF.set_override('enable_prometheus_exporter', True, group='polling')
        CONF.set_override('prometheus_listen_addresses',
                          ['127.0.0.1:9101'], group='polling')
        CONF.set_override('prometheus_tls_enable', True, group='polling')
        CONF.set_override('prometheus_tls_certfile', "cert.pem",
                          group='polling')
        # prometheus_tls_keyfile defaults to None (missing key)
        self.assertRaises(ValueError, manager.AgentManager, 0, CONF)

    def test_collect_metrics(self):
        prom_exporter.collect_metrics(self.test_image_size)
        sample_dict_1 = {'counter': 'image.size',
                         'image': 'f9276c96-8a12-432b-96a1-559d70715f97',
                         'project': 'd965489b7f894cbda89cd2e25bfd85a0',
                         'publisher': 'ceilometer',
                         'resource': 'f9276c96-8a12-432b-96a1-559d70715f97',
                         'resource_name': 'cirros2',
                         'type': 'size',
                         'unit': 'B',
                         'server_group': 'server_group123'}
        self.assertEqual(16344576,
                         prom_exporter.CEILOMETER_REGISTRY.
                         get_sample_value('ceilometer_image_size',
                                          sample_dict_1))

        prom_exporter.collect_metrics(self.test_memory_usage)
        sample_dict_2 = {'counter': 'memory.usage',
                         'memory': 'e536fff6-b20d-4aa5-ac2f-d15ac8b3af63',
                         'project': 'd965489b7f894cbda89cd2e25bfd85a0',
                         'publisher': 'ceilometer',
                         'resource': 'e536fff6-b20d-4aa5-ac2f-d15ac8b3af63',
                         'resource_name': 'myserver:instance-00000002',
                         'type': 'usage',
                         'unit': 'MiB',
                         'user': '6e7d71415cd5401cbe103829c9c5dec2',
                         'vm_instance': 'e0d297f5df3b62ec73c8d42b',
                         'server_group': 'none',
                         'flavor_id': '4af9ac72-5787-4f86-8644-0faa87ce7c83',
                         'flavor_name': 'tiny'}
        self.assertEqual(37.98046875,
                         prom_exporter.CEILOMETER_REGISTRY.
                         get_sample_value('ceilometer_memory_usage',
                                          sample_dict_2))

        prom_exporter.collect_metrics(self.test_disk_latency)
        sample_dict_3 = {'counter': 'disk.device.read.latency',
                         'disk': 'read',
                         'project': 'd965489b7f894cbda89cd2e25bfd85a0',
                         'publisher': 'ceilometer',
                         'resource':
                         'e536fff6-b20d-4aa5-ac2f-d15ac8b3af63-vda',
                         'resource_name': 'myserver:instance-00000002',
                         'type': 'device',
                         'unit': 'ns',
                         'user': '6e7d71415cd5401cbe103829c9c5dec2',
                         'vm_instance': 'e0d297f5df3b62ec73c8d42b',
                         'server_group': 'none',
                         'flavor_id': '4af9ac72-5787-4f86-8644-0faa87ce7c83',
                         'flavor_name': 'tiny'}
        # The value has to be of the second sample, as this is now a Gauge
        self.assertEqual(232128754,
                         prom_exporter.CEILOMETER_REGISTRY.
                         get_sample_value(
                             'ceilometer_disk_device_read_latency',
                             sample_dict_3))

    def test_gen_labels(self):
        slabels1 = dict(keys=[], values=[])
        slabels1['keys'] = ['disk', 'publisher', 'type', 'counter',
                            'project', 'user', 'unit', 'resource',
                            'vm_instance', 'resource_name',
                            'server_group', 'flavor_id', 'flavor_name']
        slabels1['values'] = ['read', 'ceilometer', 'device',
                              'disk.device.read.latency',
                              'd965489b7f894cbda89cd2e25bfd85a0',
                              '6e7d71415cd5401cbe103829c9c5dec2',
                              'ns',
                              'e536fff6-b20d-4aa5-ac2f-d15ac8b3af63-vda',
                              'e0d297f5df3b62ec73c8d42b',
                              'myserver:instance-00000002', 'none',
                              '4af9ac72-5787-4f86-8644-0faa87ce7c83',
                              'tiny']
        label1 = prom_exporter._gen_labels(self.test_disk_latency[0])
        self.assertDictEqual(label1, slabels1)

        slabels2 = dict(keys=[], values=[])
        slabels2['keys'] = ['memory', 'publisher', 'type', 'counter',
                            'project', 'user', 'unit', 'resource',
                            'vm_instance', 'resource_name',
                            'server_group', 'flavor_id', 'flavor_name']
        slabels2['values'] = ['e536fff6-b20d-4aa5-ac2f-d15ac8b3af63',
                              'ceilometer', 'usage', 'memory.usage',
                              'd965489b7f894cbda89cd2e25bfd85a0',
                              '6e7d71415cd5401cbe103829c9c5dec2', 'MiB',
                              'e536fff6-b20d-4aa5-ac2f-d15ac8b3af63',
                              'e0d297f5df3b62ec73c8d42b',
                              'myserver:instance-00000002', 'none',
                              '4af9ac72-5787-4f86-8644-0faa87ce7c83',
                              'tiny']
        label2 = prom_exporter._gen_labels(self.test_memory_usage[0])
        self.assertDictEqual(label2, slabels2)

        slabels3 = dict(keys=[], values=[])
        slabels3['keys'] = ['image', 'publisher', 'type', 'counter',
                            'project', 'unit', 'resource',
                            'resource_name', 'server_group']
        slabels3['values'] = ['f9276c96-8a12-432b-96a1-559d70715f97',
                              'ceilometer', 'size', 'image.size',
                              'd965489b7f894cbda89cd2e25bfd85a0', 'B',
                              'f9276c96-8a12-432b-96a1-559d70715f97',
                              'cirros2', 'server_group123']
        label3 = prom_exporter._gen_labels(self.test_image_size[0])
        self.assertDictEqual(label3, slabels3)

    @mock.patch.object(prom_exporter.CEILOMETER_REGISTRY, 'unregister')
    def test_purge_stale_metrics_existing_metric(self, mock_unregister):
        mock_metric = mock.MagicMock()
        prom_exporter.CEILOMETER_REGISTRY._names_to_collectors = {
            'ceilometer_test_metric': mock_metric
        }

        prom_exporter.purge_stale_metrics('test.metric')

        mock_unregister.assert_called_once_with(mock_metric)

    @mock.patch.object(prom_exporter.CEILOMETER_REGISTRY, 'unregister')
    def test_purge_stale_metrics_no_existing_metric(self, mock_unregister):
        prom_exporter.CEILOMETER_REGISTRY._names_to_collectors = {}

        prom_exporter.purge_stale_metrics('nonexistent.metric')

        mock_unregister.assert_not_called()

    @mock.patch.object(prom_exporter.CEILOMETER_REGISTRY, 'unregister')
    def test_purge_stale_metrics_name_transformation(self, mock_unregister):
        mock_metric = mock.MagicMock()
        prom_exporter.CEILOMETER_REGISTRY._names_to_collectors = {
            'ceilometer_cpu_util': mock_metric
        }

        prom_exporter.purge_stale_metrics('cpu.util')

        mock_unregister.assert_called_once_with(mock_metric)



#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_log import log

from ceilometer.polling import plugin_base
from ceilometer import sample

LOG = log.getLogger(__name__)

# Octavia operating_status values mapped to numeric values
OPERATING_STATUS = {
    'online': 1,
    'draining': 2,
    'offline': 3,
    'degraded': 4,
    'error': 5,
    'no_monitor': 6,
}

# Octavia provisioning_status values mapped to numeric values
PROVISIONING_STATUS = {
    'active': 1,
    'deleted': 2,
    'error': 3,
    'pending_create': 4,
    'pending_update': 5,
    'pending_delete': 6,
}


class _BaseLoadBalancerPollster(plugin_base.PollsterBase):
    """Base pollster for Octavia load balancer metrics."""

    FIELDS = ['name',
              'availability_zone',
              'vip_address',
              'vip_port_id',
              'provisioning_status',
              'operating_status',
              'provider',
              'flavor_id',
              ]

    @property
    def default_discovery(self):
        return 'lb_services'

    @staticmethod
    def extract_metadata(lb):
        return {k: getattr(lb, k, None)
                for k in _BaseLoadBalancerPollster.FIELDS}


class LoadBalancerOperatingStatusPollster(_BaseLoadBalancerPollster):
    """Pollster for Octavia load balancer operating status."""

    @staticmethod
    def get_status_id(value):
        if not value:
            return -1
        status = value.lower()
        return OPERATING_STATUS.get(status, -1)

    def get_samples(self, manager, cache, resources):
        for lb in resources or []:
            LOG.debug("LOAD BALANCER: %s", lb)
            status = self.get_status_id(lb.operating_status)
            if status == -1:
                LOG.warning(
                    "Unknown operating status %(status)s for load balancer "
                    "%(name)s (%(id)s), setting volume to -1",
                    {"status": lb.operating_status,
                     "name": lb.name,
                     "id": lb.id})
            yield sample.Sample(
                name='loadbalancer.operating',
                type=sample.TYPE_GAUGE,
                unit='status',
                volume=status,
                user_id=None,
                project_id=lb.project_id,
                resource_id=lb.id,
                resource_metadata=self.extract_metadata(lb)
            )


class LoadBalancerProvisioningStatusPollster(_BaseLoadBalancerPollster):
    """Pollster for Octavia load balancer provisioning status."""

    @staticmethod
    def get_status_id(value):
        if not value:
            return -1
        status = value.lower()
        return PROVISIONING_STATUS.get(status, -1)

    def get_samples(self, manager, cache, resources):
        for lb in resources or []:
            LOG.debug("LOAD BALANCER: %s", lb)
            status = self.get_status_id(lb.provisioning_status)
            if status == -1:
                LOG.warning(
                    "Unknown provisioning status %(status)s for load "
                    "balancer %(name)s (%(id)s), setting volume to -1",
                    {"status": lb.provisioning_status,
                     "name": lb.name,
                     "id": lb.id})
            yield sample.Sample(
                name='loadbalancer.provisioning',
                type=sample.TYPE_GAUGE,
                unit='status',
                volume=status,
                user_id=None,
                project_id=lb.project_id,
                resource_id=lb.id,
                resource_metadata=self.extract_metadata(lb)
            )



# Copyright 2016 Sungard Availability Services
# Copyright 2016 Red Hat
# All Rights Reserved
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from unittest import mock

import fixtures
from oslotest import base

from ceilometer.network import floatingip
from ceilometer.network.services import discovery
from ceilometer.polling import manager
from ceilometer.polling import plugin_base
from ceilometer import service


class _BaseTestFloatingIPPollster(base.BaseTestCase):

    def setUp(self):
        super().setUp()
        self.CONF = service.prepare_service([], [])
        self.manager = manager.AgentManager(0, self.CONF)
        plugin_base._get_keystone = mock.Mock()


class TestFloatingIPPollster(_BaseTestFloatingIPPollster):

    def setUp(self):
        super().setUp()
        self.pollster = floatingip.FloatingIPPollster(self.CONF)
        self.fake_fip = self.fake_get_fip_service()
        self.useFixture(fixtures.MockPatch('ceilometer.neutron_client.Client.'
                                           'fip_get_all',
                                           return_value=self.fake_fip))

    @staticmethod
    def fake_get_fip_service():
        return [{'router_id': 'e24f8a37-1bb7-49e4-833c-049bb21986d2',
                 'status': 'ACTIVE',
                 'tenant_id': '54a00c50ee4c4396b2f8dc220a2bed57',
                 'floating_network_id':
                     'f41f399e-d63e-47c6-9a19-21c4e4fbbba0',
                 'fixed_ip_address': '10.0.0.6',
                 'floating_ip_address': '65.79.162.11',
                 'port_id': '93a0d2c7-a397-444c-9d75-d2ac89b6f209',
                 'id': '18ca27bf-72bc-40c8-9c13-414d564ea367'},
                {'router_id': 'astf8a37-1bb7-49e4-833c-049bb21986d2',
                 'status': 'DOWN',
                 'tenant_id': '34a00c50ee4c4396b2f8dc220a2bed57',
                 'floating_network_id':
                     'gh1f399e-d63e-47c6-9a19-21c4e4fbbba0',
                 'fixed_ip_address': '10.0.0.7',
                 'floating_ip_address': '65.79.162.12',
                 'port_id': '453a0d2c7-a397-444c-9d75-d2ac89b6f209',
                 'id': 'jkca27bf-72bc-40c8-9c13-414d564ea367'},
                {'router_id': 'e2478937-1bb7-49e4-833c-049bb21986d2',
                 'status': 'error',
                 'tenant_id': '54a0gggg50ee4c4396b2f8dc220a2bed57',
                 'floating_network_id':
                     'po1f399e-d63e-47c6-9a19-21c4e4fbbba0',
                 'fixed_ip_address': '10.0.0.8',
                 'floating_ip_address': '65.79.162.13',
                 'port_id': '67a0d2c7-a397-444c-9d75-d2ac89b6f209',
                 'id': '90ca27bf-72bc-40c8-9c13-414d564ea367'},
                {'router_id': 'a27ac630-939f-4e2e-bbc3-09a6b4f19a77',
                 'status': 'UNKNOWN',
                 'tenant_id': '54a0gggg50ee4c4396b2f8dc220a2bed57',
                 'floating_network_id':
                     '4d0c3f4f-79c7-40ff-9b0d-6e3a396547db',
                 'fixed_ip_address': '10.0.0.9',
                 'floating_ip_address': '65.79.162.14',
                 'port_id': '59cc6efa-7c89-4730-b051-b15f594e6728',
                 'id': 'a8a11884-7666-4f35-901e-dbb84e7111b5'},
                {'router_id': '7eb0adde-6c3b-4a77-9714-f718a17afb83',
                 'status': None,
                 'tenant_id': '54a0gggg50ee4c4396b2f8dc220a2bed57',
                 'floating_network_id':
                     'bd6290e6-b014-4cd3-91f0-7e8a1b4c26ab',
                 'fixed_ip_address': '10.0.0.10',
                 'floating_ip_address': '65.79.162.15',
                 'port_id': 'd3b9436d-4b2b-4832-852b-34df7513c935',
                 'id': '27c539ca-94ce-42fc-a639-1bf2c8690d76'}]

    def test_fip_get_samples(self):
        samples = list(self.pollster.get_samples(
                       self.manager, {},
                       resources=self.fake_fip))
        self.assertEqual(len(self.fake_fip), len(samples))
        self.assertEqual({fip['id'] for fip in self.fake_fip},
                         {sample.resource_id for sample in samples})
        samples_dict = {sample.resource_id: sample for sample in samples}
        for fip in self.fake_fip:
            sample = samples_dict[fip['id']]
            for field in self.pollster.FIELDS:
                self.assertEqual(fip[field],
                                 sample.resource_metadata[field])

    def test_fip_volume(self):
        samples = list(self.pollster.get_samples(
                       self.manager, {},
                       resources=self.fake_fip))
        self.assertEqual(1, samples[0].volume)
        self.assertEqual(3, samples[1].volume)
        self.assertEqual(7, samples[2].volume)
        self.assertEqual(-1, samples[3].volume)
        self.assertEqual(-1, samples[4].volume)

    def test_get_fip_meter_names(self):
        samples = list(self.pollster.get_samples(
                       self.manager, {},
                       resources=self.fake_fip))
        self.assertEqual({'ip.floating'},
                         {s.name for s in samples})

    def test_fip_discovery(self):
        discovered_fips = discovery.FloatingIPDiscovery(
            self.CONF).discover(self.manager)
        self.assertEqual(len(self.fake_fip), len(discovered_fips))
        for fip in self.fake_fip:
            self.assertIn(fip, discovered_fips)
